SystemOrganization addCategory: #ZTimestamp!

Magnitude subclass: #ZTimestamp
	instanceVariableNames: 'jdn secs'
	classVariableNames: 'ClockEpoch ClockTimezoneOffset UnixEpoch Zero'
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestamp commentStamp: 'SvenVanCaekenberghe 4/23/2012 10:46' prior: 0!
I am ZTimestamp.
I am a Magnitude.
I represent a point in time, a combination of a date and a time.

I am an alternative for DateAndTime and TimeStamp.
I have second precision and live in the UTC/GMT/Zulu timezone.
I use ISO/International conventions and protocols only. 
I support some essential arithmetic.

I have an efficient internal representation:

	jnd - the julian day number <SmallInteger>
	secs - the number of seconds since midnight, the beginning of the day <SmallInteger>

Examples:

	ZTimestamp now.
	ZTimestamp fromString: '1969-07-20T20:17:40Z'.

There is some compatibility with existing, standard Chronology objects.
I correctly parse representations with a timezone designator
and can print a representation in arbitrary timezones. 

	(ZTimestamp fromString: DateAndTime now truncated printString) localPrintString.!

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 10/14/2016 15:31'!
@ stringOrStream
	"Shortcut/convenience method to create a ZTimetamp instance by parsing a string or stream"
	
	^ self readFrom: stringOrStream readStream! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 5/29/2013 17:06'!
clockEpoch
	"Return the epoch of the clock that we use as an instance (cached), see #now
	This is expressed in UTC regardless of #clockSeconds, see #clockTimezoneOffset"
	
	^ ClockEpoch ifNil: [
		ClockEpoch := self year: 1901 month: 1 day: 1 ]! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 5/29/2013 17:07'!
clockSeconds
	"Return the total seconds count of the clock that we use as an Integer since the #clockEpoch, 
	see #now. This can be in UTC or not, then it will be compensated using #clockTimezoneOffset"

	^ Time totalSeconds! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 5/29/2013 17:16'!
clockTimezoneOffset
	"Return the timezone offset of the clock that we use as an Integer seconds (cached), see #now.
	In a 3.0 image, which implements #primUTCMicrosecondsClock, Time>>#totalSeconds is UTC,
	where in 2.0 it is in local time."

	^ ClockTimezoneOffset
		ifNil: [ 
			ClockTimezoneOffset := (Time respondsTo: #primUTCMicrosecondsClock)
				ifTrue: [ 0 ]
				ifFalse: [ DateAndTime localTimeZone offset asSeconds ] ]! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:23'!
epoch
	"Return the epoch of my instances, the point in time from which they start counting, as an instance"
	
	^ self zero! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 4/24/2012 17:30'!
fromClockSeconds: seconds
	"Create and return a new instance from a specific number of seconds since our clock epoch, see #now"
	
	^ self new
		jdn: self clockEpoch julianDayNumber + (seconds // 86400)
		secs: self clockEpoch secondsSinceMidnight + (seconds \\ 86400)
	! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 4/21/2012 20:03'!
fromJulianDate: julianDate
	"Create and return a new ZTimestamp instance from the scientific Julian Date, a Float number of Julian Days"
	
	^ self new
		jdn: julianDate truncated
		secs: (julianDate fractionPart * 86400) truncated
		! !

!ZTimestamp class methodsFor: 'ston' stamp: 'SvenVanCaekenberghe 4/28/2012 08:37'!
fromSton: stonReader
	"Read our ISO representation"
	
	^ self readFrom: stonReader parseListSingleton readStream! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 4/24/2012 17:27'!
fromString: string
	"Parse and return an instance from a flexible ISO 8601 string representation, see #readFrom:" 

	^ self readFrom: string readStream! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 11/1/2012 13:40'!
fromString: string format: formatSpecification
	"Parse and return an instance from the string representation 
	according to formatSpecification, see #readFrom:format:" 

	^ self readFrom: string readStream format: formatSpecification! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 4/21/2012 20:04'!
fromUnixTime: integer
	"Create and return a new ZTimestamp from the integer number of seconds since the Unix or POSIX epoch"
	
	^ self unixEpoch + integer seconds
! !

!ZTimestamp class methodsFor: 'class initialization' stamp: 'SvenVanCaekenberghe 9/14/2012 17:02'!
initialize
	Smalltalk addToStartUpList: self.! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/21/2012 20:31'!
jdnFromYear: year month: month day: day
	"Compute and return the Julian Day Number from the Gregorian Calender date described by year, month and day.
	See http://en.wikipedia.org/wiki/Julian_date#Converting_Gregorian_calendar_date_to_Julian_Day_Number"
	 
	| a y m |
	a := (14 - month) // 12.
	y := year + 4800 - a.
	m := month + (12 * a) - 3.
	^ day 
		+ (((153 * m) + 2) // 5) 
		+ (365 * y)
		+ (y // 4)
		- (y // 100)
		+ (y // 400)
		- 32045! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:25'!
localTimeZone
	"Return the current machine's local timezone offset as a Duration"
	
	^ DateAndTime localTimeZone! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 4/21/2012 20:04'!
now
	"Create and return a new ZTimestamp equal to the current system time"
	
	^ self fromClockSeconds: (self clockSeconds - self clockTimezoneOffset)! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 2/25/2016 22:31'!
readFrom: stream
	"Parse a ISO 8601 representation from stream, in the format
	year[-]month[-]day[T]hour[:]minute[:]second[[+|-]hh[:][mm]]
	where the whole part after T as well as the TZ offset is optional.
	TZ offset can be Z, all separators can be any non digit"

	| negative year month day hour minute second offset |
	negative := stream peekFor: $-.
	year := self readInteger: 4 from: stream.
	negative ifTrue: [ year := year negated ].
	self readOptionalSeparatorFrom: stream.
	month := self readInteger: 2 from: stream.
	self readOptionalSeparatorFrom: stream.
	day := self readInteger: 2 from: stream.
	self readOptionalSeparatorFrom: stream.
	stream atEnd
		ifFalse: [
			hour := self readInteger: 2 from: stream.
			self readOptionalSeparatorFrom: stream.
			minute := self readInteger: 2 from: stream.
			self readOptionalSeparatorFrom: stream.
			second := self readInteger: 2 from: stream.
			offset := self readTimezoneOffsetFrom: stream ]
		ifTrue: [
			hour := minute := second := offset := 0 ].
	^ self 
		year: year
		month: month
		day: day
		hour: hour
		minute: minute
		second: second
		offset: (offset = 0 ifTrue: [ 0 ] ifFalse: [ offset seconds ])! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 11/1/2012 13:43'!
readFrom: stream format: formatSpecification
	"Parse a representation from stream, according to formatSpecification
	and return an instance of me. See ZTimestampFormat for details."

	^ (ZTimestampFormat fromString: formatSpecification)
			readFrom: stream ! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 6/27/2013 14:02'!
readInteger: numberOfDigits from: stream
	"Parse and return a decimal number of numberOfDigits from stream.
	Fail if that is not possible"

	| integer |
	integer := 0.
	numberOfDigits timesRepeat: [ | char |
		char := stream next.
		('0123456789' includes: char)
			ifTrue: [ integer := (integer * 10) + char digitValue ]
			ifFalse: [ self error: 'Decimal digit expected' ] ].
	^ integer! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2016 22:36'!
readOptionalSeparatorFrom: stream
	"Read an optional separator (non decimal digit) from stream and return it.
	Return nil if nothing was read"

	^ ('0123456789' includes: stream peek)
		ifTrue: [ nil ]
		ifFalse: [ stream next ]! !

!ZTimestamp class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2016 22:37'!
readTimezoneOffsetFrom: stream
	"Read and return an optional timezone offset in the form of 
	[+|-]hh[separator][mm] from stream as an Integer seconds.
	If there is no offset, just return 0"

	| sign hour minute |
	(stream atEnd or: [ stream peekFor: $Z ]) ifTrue: [ ^ 0 ].
	^ ('+-' includes: stream peek)
		ifTrue: [
			sign := stream next = $- ifTrue: [ -1 ] ifFalse: [ 1 ].
			hour := self readInteger: 2 from: stream.
			((self readOptionalSeparatorFrom: stream) notNil
					or: [ '0123456789' includes: stream peek ])
				ifTrue: [ minute := self readInteger: 2 from: stream ]
				ifFalse: [ minute := 0 ].
			sign * ((hour * 3600) + (minute * 60)) ]
		ifFalse: [ 0 ]
! !

!ZTimestamp class methodsFor: 'system startup' stamp: 'SvenVanCaekenberghe 9/14/2012 17:01'!
startUp
	"Clear the cached clock timezone offset"
	
	ClockTimezoneOffset := nil! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 10/14/2016 15:31'!
today
	"Create and return a new ZTimestamp equal to the current date"
	
	^ self now asBeginOfDay! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/18/2012 10:38'!
unixEpoch
	"Return the Unix or POSIX epoch as a ZTimestamp"
	
	^ UnixEpoch 
		ifNil: [ UnixEpoch := self year: 1970 month: 1 day: 1 ]! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 9/3/2012 19:20'!
withJdn: jdn dayMonthYearDo: block
	"Return the value of executing block with the Gregorian Calender day, month and year as arguments,
	as computed from my Julian Day Number, jdn.
	See http://en.wikipedia.org/wiki/Julian_date#Gregorian_calendar_from_Julian_day_number"
	
	| j g dg c dc b db a da y m d |
	j := jdn + 32044.
	g := j // 146097.
	dg := j \\ 146097.
	c := ((dg // 36524) + 1) * 3 // 4.
	dc := dg - (c * 36524).
	b := dc // 1461.
	db := dc \\ 1461.
	a := ((db // 365) + 1) * 3 // 4.
	da := db - (a * 365).
	y := (g * 400) + (c * 100) + (b * 4) + a.
	m := ((((da * 5) + 308)) // 153) - 2.
	d := da - ((m + 4) * 153 // 5) + 122.
	^ block
		value: d + 1
		value: ((m + 2) \\ 12) + 1
		value: (y - 4800 + ((m + 2) // 12)) ! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/2/2014 20:44'!
withNowDo: block
	"Evaluate block with 2 arguments: ZTimestamp now and the residual microseconds.
	Return the value of block. Works only on Pharo 3.0 and higher."
	
	"self withNowDo: [ :timestamp :microseconds | { timestamp. microseconds } ]"
	
	| microsecondClockValue totalSeconds |
	microsecondClockValue := Time microsecondClockValue.
	totalSeconds := microsecondClockValue // 1e6.
	^ block 
		cull: (self fromClockSeconds: totalSeconds - self clockTimezoneOffset)
		cull: microsecondClockValue - (totalSeconds * 1e6)! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 4/21/2012 20:06'!
year: year month: month day: day
	"Create and return a new ZTimestamp for the Gregorian calendar date described by year, month and date"
	
	^ self new
		jdn: (self jdnFromYear: year month: month day: day)
		secs: 0! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 2/25/2014 00:15'!
year: year month: month day: day hour: hour minute: minute second: second
	"Create and return a new ZTimestamp for the Gregorian calendar date described by year, month and date
	and the Civil time of day on that date described by hour, minute and second"

	^ self new
		jdn: (self jdnFromYear: year month: month day: day)
		secs: (hour * 3600) + (minute * 60) + second! !

!ZTimestamp class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 2/25/2014 00:15'!
year: year month: month day: day hour: hour minute: minute second: second offset: offset
	"Create and return a new ZTimestamp for the Gregorian calendar date described by year, month and date
	and the Civil time of day on that date described by hour, minute and second.
	If offset is non zero, the result is shifted accordingly."

	| timestamp |
	(timestamp := self new)
		jdn: (self jdnFromYear: year month: month day: day)
		secs: (hour * 3600) + (minute * 60) + second.
	offset isZero 
		ifFalse: [ timestamp := timestamp - offset ].
	^ timestamp ! !

!ZTimestamp class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/27/2013 13:57'!
zero
	"Return my zero instance, the point in time from which we start counting, see #epoch"

	^ Zero 
		ifNil: [ 
			Zero :=  self new jdn: 0 secs: 0 ]! !

!ZTimestamp methodsFor: 'arithmetic' stamp: 'SvenVanCaekenberghe 4/21/2012 19:54'!
+ operand
	"Return a new ZTimestamp that is the result of adding
	the operand considered as a Duration to the receiver"
	
	^ self class new
		jdn: jdn
		secs: secs + operand asDuration asSeconds! !

!ZTimestamp methodsFor: 'arithmetic' stamp: 'SvenVanCaekenberghe 3/28/2014 16:58'!
- operand
	"For ZTimestamp compatible operands, return a new Duration 
	equal to the difference between the reveiver and the operand.
	Else return a new ZTimestamp that is the result of subtracting 
	the operand considered as a Duration from the receiver.
	Implemented using double dispatch."
	
	^ operand subtractFromTimestamp: self! !

!ZTimestamp methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/28/2014 10:30'!
< aZTimestamp
	"Answer true when aZTimestamp is strictly later than me, false otherwise.
	Part of the Magnitude protocol"
	
	^ jdn = aZTimestamp julianDayNumber
		ifTrue: [ secs < aZTimestamp secondsSinceMidnight ]
		ifFalse: [ jdn < aZTimestamp julianDayNumber ]! !

!ZTimestamp methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/28/2014 10:30'!
= aZTimestamp
	"Answer true when aZTimestamp equals the receiver, false otherwise.
	Part of the Magnitude protocol"
	
	self == aZTimestamp ifTrue: [ ^ true ].
	self class = aZTimestamp class ifFalse: [ ^ false ].
	^ jdn = aZTimestamp julianDayNumber
		and: [ secs = aZTimestamp secondsSinceMidnight ]! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 7/6/2012 16:52'!
asBeginOfDay
	"Return a timestamp at the begin of the day that I represent"
	
	^ self class new
		jdn: jdn secs: 0 ! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:13'!
asDate
	"Convert the receiver to a Date"
	
	^ self asDateAndTime asDate! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 6/19/2012 08:57'!
asDateAndTime 
	"Convert the receiver to a DateAndTime.
	Note that I am in the UTC/GMT/Zulu timezone and have second precision"
	
	^ DateAndTime new
		setJdn: jdn 
		seconds: secs 
		nano: 0 
		offset: self offset! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 7/6/2012 16:52'!
asEndOfDay
	"Return a timestamp at the end of the day that I represent"
	
	^ self class new
		jdn: jdn secs: 86399! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:15'!
asMonth
	"Convert the receiver to a Month"
	
	^ self asDateAndTime asMonth! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:16'!
asTime
	"Convert the receiver to a Time in the current day.
	Note that I have second precision"
	
	^ Time seconds: secs! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/21/2012 19:59'!
asUnixTime
	"Return the Integer number of seconds since the Unix or POSIX epoch."
	
	^ (self - self class unixEpoch) asSeconds! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:16'!
asWeek
	"Convert the receiver to a Week"
	
	^ self asDateAndTime asWeek! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:17'!
asYear
	"Convert the receiver to a Year"
	
	^ self asDateAndTime asYear! !

!ZTimestamp methodsFor: 'converting' stamp: 'SvenVanCaekenberghe 3/28/2014 17:09'!
asZTimestamp
	"Convert the receiver to a ZTimestamp."
	
	^ self! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 16:57'!
day
	"Return the day of the year of the receiver, according to the Gregorian calendar, a number between 1 and 365"
	
	^ self dayOfYear! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 16:58'!
dayMonthYearDo: block
	"Compute and return the Julian Day Number from the Gregorian Calender date described by year, month and day.
	See http://en.wikipedia.org/wiki/Julian_date#Converting_Gregorian_calendar_date_to_Julian_Day_Number"

	^ self class 
		withJdn: jdn dayMonthYearDo: block! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 16:59'!
dayOfMonth
	"Return the day of the month of the receiver, according to the Gregorian calendar, a number between 1 and 31"

	^ self dayMonthYearDo: [ :day :month :year | day ]! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/1/2014 15:35'!
dayOfWeek
	"Return the day of the week of the receiver, according to the Gregorian calendar, a number between 1 and 7.
	1 = Sunday, 2 = Monday, .. , 7 = Saturday."

	^ ((jdn + 1) \\ 7) + 1! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/1/2014 15:36'!
dayOfWeekName
	"Return the English name of the day of the week of the receiver, according to the Gregorian calendar, a Symbol"

	^ #(Sunday Monday Tuesday Wednesday Thursday Friday Saturday) at: self dayOfWeek! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:01'!
dayOfYear
	"Return the day of the year of the receiver, according to the Gregorian calendar, a number between 1 and 365"

	^ self dayMonthYearDo: [ :day :month :year | | monthStart |
		monthStart := #(1 32 60 91 121 152 182 213 244 274 305 335) at: month.
		(month > 2 and: [ Year isLeapYear: year ])
			ifTrue: [ monthStart + day ]
			ifFalse: [ monthStart + day - 1 ] ]! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 11/1/2012 13:45'!
format: formatSpecification
	"Return a string representation of me according to formatSpecification.
	See ZTimestampFormat for details."
	
	^ (ZTimestampFormat fromString: formatSpecification)
			format: self! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 11/1/2012 13:46'!
format: formatSpecification on: stream
	"Write a string representation of me according to formatSpecification on stream.
	See ZTimestampFormat for details."

	^ (ZTimestampFormat fromString: formatSpecification)
			format: self on: stream! !

!ZTimestamp methodsFor: 'hash' stamp: 'SvenVanCaekenberghe 5/11/2012 07:41'!
hash
	"Return an Integer hash value for the receiver.
	Part of the Magnitude protocl"
	
	^ jdn hashMultiply bitXor: secs! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:18'!
hour
	"Return the number of hours in the receiver, according to Civil time, a number between 0 and 23"

	^ (secs \\ 86400) // 3600! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:18'!
hours
	"Return the number of hours in the receiver, according to Civil Time, a number between 0 and 23"

	^ self hour! !

!ZTimestamp methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 4/24/2012 16:45'!
jdn: julianDayNumber secs: secondsSinceMidnight
	"Initialize the receiver with julianDayNumber and secondsSinceMidnight,
	then normalize so that secondsSinceMidnight is within the expected range"
	
	jdn := julianDayNumber.
	secs := secondsSinceMidnight.
	self normalize! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/21/2012 20:00'!
julianDate
	"Return the scientific Julian Date, a Float number of Julian Days"
	
	^ jdn + (secs / 86400.0)! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:05'!
julianDayNumber
	"Return the Julian Day Number of the receiver, an Integer number of days since the epoch"
	
	^ jdn! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 4/21/2012 20:09'!
localPrintString
	"Convencience method to convert the receiver in an ISO 8601 representation using the local timezone"

	^ String streamContents: [ :stream |
		self printLocalOn: stream ]! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:18'!
minute
	"Return the number of minutes in the receiver after full hours have been removed, 
	according to Civil Time, a number between 0 and 59"

	^ (secs \\ 3600) // 60! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:18'!
minutes
	"Return the number of minutes in the receiver after full hours have been removed, 
	according to Civil Time, a number between 0 and 59"

	^ self minute! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 16:58'!
month
	"Return the month of the receiver, according to the Gregorian calendar, a number between 1 and 12"

	^ self dayMonthYearDo: [ :day :month :year | month ]! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 16:58'!
monthIndex
	"Return the month of the receiver, according to the Gregorian calendar, a number between 1 and 12"

	^ self month! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/1/2014 13:01'!
monthName
	"Return the English name of the month of the receiver, 
	according to the Gregorian calendar, a Symbol"

	^ #(
		January February March 
		April May June 
		July August September 
		October November December ) at: self monthIndex! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 19:43'!
nanoSecond
	"For compatibility: make clear that I have second precision and no fractions of a second."
	 
	^ 0! !

!ZTimestamp methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 4/22/2012 18:55'!
normalize
	"Normalize the receiver such that 0 <= secs < 86400"
	 
	(secs between: 0 and: 86399)
		ifFalse: [
			jdn := jdn + (secs // 86400).
			secs := secs \\ 86400 ]! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:14'!
offset
	"For compatibility: make clear that my timezone offset is zero.
	I am always in the UTC, GMT or Zulu timezone"
	 
	^ Duration zero! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 4/21/2012 19:45'!
printLocalOn: stream
	"Print the receiver on stream in ISO 8601 extended format in the local timezone"
	
	self 
		printOn: stream 
		format: #($- $T $: nil) 
		timezone: self class localTimeZone offset! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:19'!
printOn: stream
	"Print the receiver on stream in ISO 8601 extended format"
	
	self 
		printOn: stream 
		format: #($- $T $: $Z)! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 5/28/2014 21:35'!
printOn: stream format: format
	"Print the receiver on stream using format, following ISO 8601.
	Format defines 4 characters to use:
	#(dateSeparator dateTimeSeparator timeSeparator timeZoneIndicator)
	Each can be nil to skip that separator"
	
	self dayMonthYearDo: [ :day :month :year |
		year < 0
			ifTrue: [ 
				stream nextPut: $-. 
				year abs printOn: stream base: 10 length: 4 padded: true  ]
			ifFalse: [ 
				year printOn: stream base: 10 length: 4 padded: true ].
		(format at: 1) ifNotNil: [ :separator | stream nextPut: separator ].
		month printOn: stream base: 10 length: 2 padded: true.
		(format at: 1) ifNotNil: [ :separator | stream nextPut: separator ].
		day printOn: stream base: 10 length: 2 padded: true ].
	(format at: 2) ifNotNil: [ :separator | stream nextPut: separator ].
	self hour printOn: stream base: 10 length: 2 padded: true.
	(format at: 3) ifNotNil: [ :separator | stream nextPut: separator ].
	self minute printOn: stream base: 10 length: 2 padded: true.
	(format at: 3) ifNotNil: [ :separator | stream nextPut: separator ].
	self second printOn: stream base: 10 length: 2 padded: true.
	(format at: 4) ifNotNil: [ :separator | stream nextPut: separator ]! !

!ZTimestamp methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 6/27/2013 13:54'!
printOn: stream format: format timezone: timezone
	"Print an ISO representation in timezone and adding it as [+|-]HH:MM"

	(self + timezone asDuration) printOn: stream format: format.
	stream nextPut: (timezone positive ifTrue: [ $+ ] ifFalse: [ $- ]).
	timezone hours abs printOn: stream base: 10 length: 2 padded: true.
	(format at: 3) ifNotNil: [ :separator | stream nextPut: separator ].
	timezone minutes printOn: stream base: 10 length: 2 padded: true! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:17'!
second
	"Return the number of seconds in the receiver after full hours and minutes have been removed, 
	according to Civil Time, a number between 0 and 59"

	^ secs \\ 60! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2014 00:17'!
seconds
	"Return the number of seconds in the receiver after full hours and minutes have been removed, 
	according to Civil Time, a number between 0 and 59"

	^ self second! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:09'!
secondsSinceMidnight
	"Return the seconds since midnight since the beginning of the Julian Day of the receiver, 
	an Integer between 0 and 86399, inclusive"
	
	^ secs! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 7/10/2014 13:55'!
secondsSinceMidnightUTC
	"Return the seconds since midnight since the beginning of the Julian Day of the receiver, 
	an Integer between 0 and 86399, inclusive - we're always in UTC"
	
	^ secs! !

!ZTimestamp methodsFor: 'ston' stamp: 'SvenVanCaekenberghe 11/29/2012 16:27'!
stonContainSubObjects 
	^ false! !

!ZTimestamp methodsFor: 'ston' stamp: 'SvenVanCaekenberghe 4/28/2012 09:30'!
stonOn: stonWriter
	"Use an ISO representation with all details"
	
	stonWriter writeObject: self listSingleton: 
		(String new: 20 streamContents: [ :stream |
			self printOn: stream ])! !

!ZTimestamp methodsFor: 'arithmetic' stamp: 'SvenVanCaekenberghe 3/28/2014 17:07'!
subtractDuration: duration
	"Part of the double dispatch implementation of ZTimestamp>>#-"
	
	^ self + duration negated! !

!ZTimestamp methodsFor: 'arithmetic' stamp: 'SvenVanCaekenberghe 3/28/2014 17:07'!
subtractFromTimestamp: timestamp
	"Part of the double dispatch implementation of ZTimestamp>>#-"
	
	^ timestamp subtractTimestamp: self! !

!ZTimestamp methodsFor: 'arithmetic' stamp: 'SvenVanCaekenberghe 3/28/2014 17:07'!
subtractTimestamp: timestamp
	"Part of the double dispatch implementation of ZTimestamp>>#-"
	
	^ Duration seconds: 
			((jdn - timestamp julianDayNumber) * 86400) 
			+ (secs - timestamp secondsSinceMidnight)! !

!ZTimestamp methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/24/2012 17:11'!
year
	"Return the year of the receiver, according to the Gregorian calendar, an Integer"

	^ self dayMonthYearDo: [ :day :month :year | year ]! !

DynamicVariable subclass: #ZCurrentTimezone
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZCurrentTimezone commentStamp: 'SvenVanCaekenberghe 3/12/2014 23:04' prior: 0!
I am ZCurrentTimezone.
I am a DynamicVariable.

I am used to hold the default timezone to use to render UTC/GMT/Zulu timestamps to local time representations.

I am also used to hold the default timezone to parse local time representations into UTC/GMT/Zulu timestamps.

Any object that responds to #offsetForTimestamp: can be a timezone.
!

!Date methodsFor: '*ztimestamp' stamp: 'SvenVanCaekenberghe 5/2/2016 10:23'!
asZTimestamp
	"Convert the receiver to a ZTimestamp.
	Note that we are ignoring timezone information 
	and interpret the receiver as an abstract data in UTC"

	^ ZTimestamp year: self year month: self monthIndex day: self dayOfMonth! !

!Duration methodsFor: '*ztimestamp' stamp: 'SvenVanCaekenberghe 3/28/2014 17:08'!
subtractFromTimestamp: timestamp
	"Part of the double dispatch implementation of ZTimestamp>>#-"
	
	^ timestamp subtractDuration: self! !

TestCase subclass: #ZTimestampFormatTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampFormatTests commentStamp: '<historical>' prior: 0!
I am ZTimestampFormatTests, a suite of unit tests for ZTimestampFormat.!

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/1/2012 19:19'!
testAMPM
	| times format12 format24 |
	times := #( 
		('12:00AM' '00:00') ('12:01AM' '00:01') ('12:59AM' '00:59') ('01:00AM' '01:00') ('11:00AM' '11:00') ('11:59AM' '11:59')
		('12:00PM' '12:00') ('12:01PM' '12:01') ('12:59PM' '12:59')('01:00PM' '13:00') ('11:00PM' '23:00') ('11:59PM' '23:59') ).
	format12 := ZTimestampFormat fromString: '04:05PM'.
	format24 := ZTimestampFormat fromString: '16:05'.
	times do: [ :each |
		self assert: (format12 parse: each first) equals: (format24 parse: each second).
		self assert: (format12 format: (format12 parse: each first)) equals: each first.
		self assert: (format24 format: (format24 parse: each second)) equals: each second ]! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/13/2014 15:33'!
testCurrentTimezoneFormatting
	| format winter summer |
	winter := ZTimestamp year: 2014 month: 1 day: 1 hour: 11 minute: 0 second: 0.
	summer := ZTimestamp year: 2014 month: 8 day: 1 hour: 11 minute: 0 second: 0.
	ZCurrentTimezone 
		value: (ZTimezone id: #'Europe/Brussels')
		during: [ 
			format := ZTimestampFormat iso.
			format useCurrentTimezone.
			self 
				assert: (format format: winter)
				equals: '2014-01-01T12:00:00+01:00'.
			self 
				assert: (format format: summer)
				equals: '2014-08-01T13:00:00+02:00'.
			format specification: '20010203T1605'.
			self 
				assert: (format format: winter)
				equals: '20140101T1200'.
			self 
				assert: (format format: summer)
				equals: '20140801T1300' ]! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/13/2014 15:53'!
testCurrentTimezoneParsing
	| format winter summer |
	winter := ZTimestamp year: 2014 month: 1 day: 1 hour: 11 minute: 0 second: 0.
	summer := ZTimestamp year: 2014 month: 8 day: 1 hour: 11 minute: 0 second: 0.
	ZCurrentTimezone 
		value: (ZTimezone id: #'Europe/Brussels')
		during: [ 
			format := ZTimestampFormat fromString: '2001-02-03T16:05'.
			format useCurrentTimezone.
			self 
				assert: (format parse: '2014-01-01T12:00')
				equals: winter.
			self 
				assert: (format parse: '2014-08-01T13:00')
				equals: summer.
			format specification: '2001-02-03T16:05+00:00'.
			self 
				assert: (format parse: '2014-01-01T16:00+05:00')
				equals: winter.
			self 
				assert: (format parse: '2014-08-01T15:00+04:00')
				equals: summer ]! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:32'!
testDateAndTime
	| now format |
	now := DateAndTime now rounded.
	format := ZTimestampFormat fromString: '2001-02-03T16:05:06+00:00'.
	self 
		assert: (format format: now)
		equals: now printString.	
	format createDateAndTime.
	self
		assert: (format parse: now printString)
		equals: now.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 6/10/2015 16:41'!
testDateAndTimeWithNanoseconds
	| now format |
	now := DateAndTime now.
	format := ZTimestampFormat fromString: '2001-02-03T16:05:06.7+00:00'.
	self 
		assert: (format format: now)
		equals: now printString.	
	format createDateAndTime.
	self
		assert: (format parse: now printString)
		equals: now.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:44'!
testDateCreation
	self 
		assert: ((ZTimestampFormat fromString: '03/02/2001') createDate; parse: '07/08/1967')
		equals: (Date year: 1967 month: 8 day: 7).	! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/1/2014 13:16'!
testDateFormatting
	self
		assert: ((ZTimestampFormat fromString: '2001/02/03') format: (ZTimestamp year: 2012 month: 10 day: 20))
		equals: '2012/10/20'.
	self
		assert: ((ZTimestampFormat fromString: '3 February 2001') format: (ZTimestamp year: 2012 month: 10 day: 20))
		equals: '20 October 2012'.
	self
		assert: ((ZTimestampFormat fromString: '_3 February 2001') 
						french;
						format: (ZTimestamp year: 1967 month: 8 day: 7))
		equals: ' 7 Août 1967'.
	self
		assert: ((ZTimestampFormat fromString: '3 FEB ''01') format: (ZTimestamp year: 2012 month: 10 day: 20))
		equals: '20 OCT ''12'.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/1/2014 13:17'!
testDateParsing
	self
		assert: ((ZTimestampFormat fromString: '2001/02/03') parse: '2012/10/20') 
		equals: (ZTimestamp year: 2012 month: 10 day: 20).
	self
		assert: ((ZTimestampFormat fromString: '3 FEB ''01') parse: '20 OCT ''12') 
		equals: (ZTimestamp year: 2012 month: 10 day: 20).
	self
		assert: ((ZTimestampFormat fromString: '3 February 2001') parse: '20 October 2012') 
		equals: (ZTimestamp year: 2012 month: 10 day: 20).
	self
		assert: ((ZTimestampFormat fromString: '_3 February 2001') 
						french;
						parse: ' 7 Août 1967') 
		equals: (ZTimestamp year: 1967 month: 8 day: 7).! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/13/2014 15:33'!
testExplicitTimezoneFormatting
	| format winter summer |
	winter := ZTimestamp year: 2014 month: 1 day: 1 hour: 11 minute: 0 second: 0.
	summer := ZTimestamp year: 2014 month: 8 day: 1 hour: 11 minute: 0 second: 0.
	format := ZTimestampFormat iso.
	format timezone: (ZTimezone id: #'Europe/Brussels').
	self 
		assert: (format format: winter)
		equals: '2014-01-01T12:00:00+01:00'.
	self 
		assert: (format format: summer)
		equals: '2014-08-01T13:00:00+02:00'.
	format specification: '20010203T1605'.
	self 
		assert: (format format: winter)
		equals: '20140101T1200'.
	self 
		assert: (format format: summer)
		equals: '20140801T1300'.
	! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/19/2012 13:33'!
testFlatDateFormatting
	self
		assert: ((ZTimestampFormat fromString: '20010203') format: (Date year: 1967 month: 8 day: 7))
		equals: '19670807'.
	self
		assert: ((ZTimestampFormat fromString: '02032001') format: (Date year: 1967 month: 8 day: 7))
		equals: '08071967'.
	self
		assert: ((ZTimestampFormat fromString: '03022001') format: (Date year: 1967 month: 8 day: 7))
		equals: '07081967'.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/1/2014 13:17'!
testMonthNames
	#(english french german dutch) do: [ :language | | format |
		format := ZTimestampFormat fromString: '03-February-2001'.
		format perform: language.
		1 to: 12 do: [ :month | | timestamp |
			timestamp := ZTimestamp year: 2012 month: month day: 10. 
			self 
				assert: (format parse: (format format: timestamp))
				equals: timestamp ] ]! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/13/2014 16:11'!
testPredefinedFormats
	| reference now |
	reference := ZTimestampFormat reference24.
	now := DateAndTime now truncated.
	#(ansiC verbose kitchen rfc3339 rfc822 rfc850 unixDate) 
		do: [ :each | | format target |
			format := ZTimestampFormat perform: each.
			target := format specification copy replaceAll: $_ with: $ . 
			self assert: (format format: reference) equals: target ].
	#(verbose rfc3339 iso isoTZ) 
		do: [ :each | | format |
			(format := ZTimestampFormat perform: each)
				createDateAndTime.
			self assert: (format parse: (format format: now)) equals: now ].
	now := ZTimestamp now.
	#(isoCompact) 
		do: [ :each | | format|
			format := ZTimestampFormat perform: each.
			self assert: (format parse: (format format: now)) equals: now ].
	! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:32'!
testReferenceDateAndTime
	| reference format now |
	reference := ZTimestampFormat reference12 asDateAndTime.
	(format := ZTimestampFormat fromString: reference printString)
		createDateAndTime.
	self assert: (format format: reference) equals: reference printString.
	self assert: (format parse: reference printString) equals: reference.
	reference := ZTimestampFormat reference24 asDateAndTime.
	(format := ZTimestampFormat fromString: reference printString)
		createDateAndTime.
	self assert: (format format: reference) equals: reference printString.
	self assert: (format parse: reference printString) equals: reference.
	now := DateAndTime now truncated.
	self assert: (format format: now) equals: now printString.
	self assert: (format parse: now printString) equals: now.
	! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/1/2012 20:39'!
testRfc3339
	| now |
	now := ZTimestamp now.
	self 
		assert: (ZTimestampFormat rfc3339 format: now)
		equals: now printString.
	self
		assert: (ZTimestampFormat rfc3339 parse: now printString)
		equals: now.
	self 
		assert: (ZTimestampFormat rfc3339 format: ZTimestampFormat reference12)
		equals: ZTimestampFormat reference12 printString.
	self
		assert: (ZTimestampFormat rfc3339 parse: ZTimestampFormat reference12 printString)
		equals: ZTimestampFormat reference12.
	self 
		assert: (ZTimestampFormat rfc3339 format: ZTimestampFormat reference24)
		equals: ZTimestampFormat reference24 printString.
	self
		assert: (ZTimestampFormat rfc3339 parse: ZTimestampFormat reference24 printString)
		equals: ZTimestampFormat reference24.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:47'!
testTimeCreation
	self 
		assert: ((ZTimestampFormat fromString: '4:05:06PM') createTime; parse: '8:11:12PM')
		equals: (Time hour: 20 minute: 11 second: 12).	! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 10/30/2012 14:43'!
testTimeFormatting
	self
		assert: ((ZTimestampFormat fromString: '04:05:06') 
						format: (ZTimestamp year: 0 month: 0 day: 0 hour: 11 minute: 30 second: 19))
		equals: '11:30:19'.! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 10/31/2012 20:40'!
testTimeParsing
	self
		assert: ((ZTimestampFormat fromString: '04:05:06') parse: '11:30:19')
		equals: (ZTimestamp year: 0 month: 0 day: 0 hour: 11 minute: 30 second: 19).
	self
		assert: ((ZTimestampFormat fromString: '16:05:06') parse: '22:30:19')
		equals: (ZTimestamp year: 0 month: 0 day: 0 hour: 22 minute: 30 second: 19).
	self
		assert: ((ZTimestampFormat fromString: '04:05:06 PM') parse: '11:30:19 AM')
		equals: (ZTimestamp year: 0 month: 0 day: 0 hour: 11 minute: 30 second: 19).
	self
		assert: ((ZTimestampFormat fromString: '04:05:06 PM') parse: '11:30:19 PM')
		equals: (ZTimestamp year: 0 month: 0 day: 0 hour: 23 minute: 30 second: 19).! !

!ZTimestampFormatTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/1/2014 15:36'!
testWeekdayNames
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 11 day: 3. 
	#(english french german dutch) do: [ :language | | format |
		format := ZTimestampFormat fromString: 'Saturday, 03-Feb-2001'.
		format perform: language.
		1 to: 7 do: [ :weekday |
			self 
				assert: (format parse: (format format: timestamp + weekday days))
				equals: (timestamp + weekday days) ] ]! !

TestCase subclass: #ZTimestampSNTPClientTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampSNTPClientTests commentStamp: '<historical>' prior: 0!
I am ZTimestampSNTPClientTests, a test suite for ZTimestampSNTPClient.!

ZTimestampSNTPClientTests subclass: #ZTimestampPreciseSNTPClientTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampPreciseSNTPClientTests commentStamp: '<historical>' prior: 0!
I am ZTimestampPreciseSNTPClientTests, a test suite for ZTimestampPreciseSNTPClient.!

!ZTimestampPreciseSNTPClientTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:08'!
clientClass
	^ ZTimestampPreciseSNTPClient ! !

!ZTimestampSNTPClientTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:07'!
clientClass
	^ ZTimestampSNTPClient ! !

!ZTimestampSNTPClientTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:07'!
testClockDifferenceMulti
	| differences |
	differences := (1 to: 10) collect: [ :each |
		| client difference |
		client := self clientClass new.
		difference := client clockDifference.
		client close.
		difference seconds ].
	self assert: differences average <= 2! !

!ZTimestampSNTPClientTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:07'!
testClockDifferenceSingle
	| differences client |
	client := self clientClass new.
	differences := (1 to: 10) collect: [ :each |
		| difference |
		difference := client clockDifference.
		difference seconds ].
	client close.
	self assert: differences average <= 2! !

!ZTimestampSNTPClientTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:07'!
testEnforceClockDifference
	self clientClass new 
		enforceClockDifference: 5 seconds 
		ifFail: [ :difference | 
			self fail: ('Clock difference too large ({1} > 5s)' format: { difference } ) ]! !

!ZTimestampSNTPClientTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/26/2013 10:21'!
testLocalClockOffset
	| client offset |
	client := self clientClass new.
	offset := client localClockOffset.
	client close.
	self assert: offset <= 2 seconds! !

!ZTimestampSNTPClientTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/26/2013 10:21'!
testRoundtripDelay
	| client delay |
	client := self clientClass new.
	delay := client roundtripDelay.
	client close.
	self assert: delay <= 2 seconds! !

TestCase subclass: #ZTimestampTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampTests commentStamp: 'SvenVanCaekenberghe 4/20/2012 16:54' prior: 0!
I am ZTimestampTests, a test suite for ZnTimestamp.!

!ZTimestampTests methodsFor: 'asserting' stamp: 'SvenVanCaekenberghe 4/24/2012 17:36'!
assertInvariants: timestamp
	self assert: (timestamp isKindOf: ZTimestamp).
	self assert: timestamp julianDayNumber isInteger.
	self assert: (timestamp secondsSinceMidnight between: 0 and: (24 * 60 * 60) - 1).
	self assert: timestamp offset = Duration zero! !

!ZTimestampTests methodsFor: 'asserting' stamp: 'SvenVanCaekenberghe 7/10/2014 14:00'!
assertTimestamp: timestamp equalsTimestamp: reference 
	self assert: reference julianDayNumber = timestamp julianDayNumber.
	"Let's try to stay compatible between different Pharo versions, it is just for testing anyway"
	[ self assert: reference secondsSinceMidnight = timestamp secondsSinceMidnight ]
		on: Deprecation, MessageNotUnderstood  
		do: [ :exception |
			"Use the new selector in Pharo 3.0 and above"
			self assert: reference secondsSinceMidnightUTC = timestamp secondsSinceMidnightUTC.  
			exception return ].
	self assert: reference year = timestamp year.
	self assert: reference month = timestamp month.
	self assert: reference dayOfMonth = timestamp dayOfMonth.
	self assert: reference hour = timestamp hour.
	self assert: reference minute = timestamp minute.
	self assert: reference second = timestamp second! !

!ZTimestampTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/22/2012 10:20'!
lunarLanding
	"First lunar landing, Apollo 11, Mare Tranquillitatis"
	
	^ ZTimestamp fromString: '1969-07-20T20:17:40Z'! !

!ZTimestampTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/22/2012 10:20'!
marsLanding
	"First soft Mars landing, Viking 1, Chryse Planitia"
	
	^ ZTimestamp fromString: '1979-07-20T11:53:06Z'! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 21:27'!
testAddingOneDay
	| t1 t2 |
	t1 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	t2 := ZTimestamp year: 2012 month: 8 day: 8 hour: 14 minute: 20 second: 11.
	self assert: t1 < t2.
	self assert: (t1 + 1 day) = t2.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 21:28'!
testAddingOneHour
	| t1 t2 |
	t1 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	t2 := ZTimestamp year: 2012 month: 8 day: 7 hour: 15 minute: 20 second: 11.
	self assert: t1 < t2.
	self assert: (t1 + 1 hour) = t2.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/21/2016 11:07'!
testAt
	self assert: ZTimestamp @ '1969-07-20T20:17:40Z' equals: self lunarLanding.
	self assert: ZTimestamp @ '19690720T201740' equals: self lunarLanding.
	self assert: ZTimestamp @ '19690720' equals: self lunarLanding asBeginOfDay! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 20:48'!
testBasicFieldAccessors
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp year = 2012.
	self assert: timestamp month = 4.
	self assert: timestamp dayOfMonth = 20.
	self assert: timestamp hour = 15.
	self assert: timestamp minute = 50.
	self assert: timestamp seconds = 10	! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 7/6/2012 17:31'!
testBeginOfDay
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp asBeginOfDay equals: (ZTimestamp year: 2012 month: 4 day: 20)! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 16:51'!
testCalculations
	| delta |
	delta := self marsLanding - self lunarLanding.
	self assert: delta > Duration zero.
	self assert: self lunarLanding + delta equals: self marsLanding.
	self assert: self marsLanding - delta equals: self lunarLanding! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 10:51'!
testComparing
	self assertInvariants: self marsLanding.
	self deny: self lunarLanding = self marsLanding.
	self assert: self lunarLanding < self marsLanding.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 6/10/2015 16:44'!
testConversions
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp asDate year equals: 2012.
	self assert: timestamp asDate monthIndex equals: 4.
	self assert: timestamp asDate dayOfMonth equals: 20.
	self 
		assert: timestamp asTime
		equals: (Time hour: 15 minute: 50 second: 10).
	self 
		assert: timestamp asDateAndTime
		equals: (DateAndTime year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10 offset: Duration zero).
	self assert: timestamp asZTimestamp = timestamp.
	self 
		assert: (DateAndTime year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10 offset: Duration zero) asZTimestamp
		equals: timestamp.
	self 
		assert: (DateAndTime year: 2012 month: 4 day: 20 hour: 19 minute: 50 second: 10 offset: 4 hours) asZTimestamp
		equals: timestamp.
! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:47'!
testCreationWithAllFields
	| reference timestamp |
	reference := DateAndTime 
		year: 2012 
		month: 4 
		day: 20 
		hour: 15 
		minute: 50 
		second: 10 
		nanoSecond: 0 
		offset: Duration zero.
	timestamp := ZTimestamp 
		year: reference year 
		month: reference month 
		day: reference dayOfMonth 
		hour: reference hour 
		minute: reference minute 
		second: reference second.
	self assertTimestamp: reference equalsTimestamp: timestamp! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:47'!
testCreationWithDateFields
	| reference timestamp |
	reference := DateAndTime 
		year: 2012 
		month: 4 
		day: 20 
		hour: 0 
		minute: 0 
		second: 0 
		nanoSecond: 0 
		offset: Duration zero.
	timestamp := ZTimestamp 
		year: reference year 
		month: reference month 
		day: reference dayOfMonth.
	self assertTimestamp: reference equalsTimestamp: timestamp! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 7/6/2012 17:32'!
testEndOfDay
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp asEndOfDay equals: (ZTimestamp year: 2012 month: 4 day: 20 hour: 23 minute: 59 second: 59)! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 9/3/2012 19:19'!
testEpoch
	| epoch |
	epoch := ZTimestamp epoch.
	self assert: epoch equals: ZTimestamp zero.
	self assert: epoch julianDayNumber equals: 0.
	self assert: epoch secondsSinceMidnight equals: 0.
	"astronomical year to proleptic Gregorian year"
	self assert: epoch year equals: -4713.
	self assert: epoch month equals: 11.
	self assert: epoch dayOfMonth equals: 24.
	! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 21:25'!
testEquality
	| t1 t2 |
	t1 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	t2 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	self assert: t1 = t2.
	self assert: ZTimestamp now = ZTimestamp now.
	self assert: ZTimestamp unixEpoch = ZTimestamp unixEpoch.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:21'!
testFormatting
	| timestamp formatter |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	formatter := [ :format | 
		String streamContents: [ :stream | 
			timestamp printOn: stream format: format ] ].
	self assert: (formatter value: #($- $T $: $Z)) = '2012-04-20T15:50:10Z'.
	self assert: (formatter value: #(nil $T nil nil)) = '20120420T155010'.
	self assert: (formatter value: #($/ $| $: $Z)) = '2012/04/20|15:50:10Z'.
	self assert: (formatter value: #($- $T $: nil)) = '2012-04-20T15:50:10'.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:07'!
testHashing
	| array set now |
	now := ZTimestamp now.
	array := { now. self lunarLanding. self marsLanding. self unixTimeBillenium }.
	set := array asSet.
	self assert: array size = set size.
	array do: [ :each |
		self assert: (set includes: each) ].
	set addAll: array.
	self assert: array size = set size.
	array do: [ :each |
		self assert: (set includes: each) ].! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 00:50'!
testJulianDate
	| now julianDate |
	now := ZTimestamp now.
	julianDate := now julianDate.
	self assert: (ZTimestamp fromJulianDate: julianDate) - now < 1 second! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/5/2012 21:23'!
testLocalPrintString
	| now timestamp reference |
	now := DateAndTime now.
	timestamp := now asUTC asZTimestamp localPrintString.
	reference := now truncated printString.
	self assert: timestamp equals: reference! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 00:45'!
testLunarLanding
	| timestamp |
	timestamp := self lunarLanding.
	self assertInvariants: timestamp.
	self assert: timestamp monthName = #July.
	self assert: timestamp dayOfWeekName = #Sunday.
	self assert: timestamp dayOfYear = 201.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 19:05'!
testMixedCalculations
	| now delta |
	now := DateAndTime now truncated asUTC.
	delta := now - self lunarLanding.
	self assert: delta > Duration zero.
	self assert: self lunarLanding + delta equals: now asZTimestamp.
	self assert: (self lunarLanding + delta) asDateAndTime equals: now.
	self assert: (now - delta) asZTimestamp equals: self lunarLanding.
	self assert: (now - delta) equals: self lunarLanding asDateAndTime! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 20:40'!
testNow
	self assertInvariants: ZTimestamp now! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 21:28'!
testOneDayDifference
	| t1 t2 |
	t1 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	t2 := ZTimestamp year: 2012 month: 8 day: 8 hour: 14 minute: 20 second: 11.
	self assert: t1 < t2.
	self assert: (t2 - t1) = 1 day! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 21:28'!
testOneHourDifference
	| t1 t2 |
	t1 := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	t2 := ZTimestamp year: 2012 month: 8 day: 7 hour: 15 minute: 20 second: 11.
	self assert: t1 < t2.
	self assert: (t2 - t1) = 1 hour.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/25/2016 22:30'!
testOptionalTimezoneParsing
	#( 
		'1969-07-20T20:17:40Z'
		'1969-07-20T20:17:40'
		'1969-07-20T20:17:40+00:00'
		'1969-07-20T20:17:40+00'
		'1969-07-20T20:17:40+0000'
		'1969-07-20T21:17:40+01:00'
		'1969-07-20T21:17:40+01'
		'1969-07-20T21:17:40+0100'
		'1969-07-20T19:17:40-01:00'
		'1969-07-20T19:17:40-01'
		'1969-07-20T19:17:40-0100'
		'1969-07-20T21:47:40+01:30'
		'1969-07-20T21:47:40+0130'
		'1969-07-20T18:47:40-01:30'
		'1969-07-20T18:47:40-0130' ) do: [ :each |
		self assert: (ZTimestamp fromString: each) equals: self lunarLanding ]! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 5/28/2014 17:20'!
testParseEpoch
	| epoch |
	epoch := ZTimestamp epoch.
	self assert: (ZTimestamp fromString: epoch asString) equals: epoch! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 7/6/2012 17:14'!
testParsing
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 8 day: 7 hour: 14 minute: 20 second: 11.
	self assert: (ZTimestamp readFrom: '2012-08-07T14:20:11Z' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '20120807T142011' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '20120807142011' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '2012-08-07T14:20:11' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '2012-08-07T14:20:11+00:00' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '2012-08-07T16:20:11+02:00' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '2012-08-07T12:20:11-02:00' readStream) = timestamp.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/21/2012 22:14'!
testParsingDates
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 8 day: 7.
	self assert: (ZTimestamp readFrom: '2012-08-07' readStream) = timestamp.
	self assert: (ZTimestamp readFrom: '20120807' readStream) = timestamp.! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 6/27/2013 14:05'!
testParsingError
	self should: [ ZTimestamp fromString: '45451-09-06T11:01:07Z' ] raise: Error! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 5/28/2014 21:36'!
testPrintString
	| timestamp |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp printString = '2012-04-20T15:50:10Z'.
	timestamp := ZTimestamp year: -12 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	self assert: timestamp printString = '-0012-04-20T15:50:10Z'! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/23/2012 08:27'!
testPrintStringCompatibility
	| dateTime timestamp |
	dateTime := DateAndTime now truncated printString.
	timestamp := (ZTimestamp fromString: dateTime) localPrintString.
	self assert: dateTime equals: timestamp! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/28/2012 09:28'!
testSTON
	Smalltalk at: #STON ifPresent: [ :ston | 
		| now |
		now := ZTimestamp now.
		self 
			assert: (ston fromString: (ston toString: now)) 
			equals: now ]! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:03'!
testSorting
	| array sorted now |
	now := ZTimestamp now.
	array := { now. self lunarLanding. self marsLanding. self unixTimeBillenium }.
	sorted := array sorted.
	self assert: sorted first = self lunarLanding.
	self assert: sorted last = now! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:35'!
testTimezoneFormatting
	| timestamp formatter |
	timestamp := ZTimestamp year: 2012 month: 4 day: 20 hour: 15 minute: 50 second: 10.
	formatter := [ :format :timezone| 
		String streamContents: [ :stream | 
			timestamp printOn: stream format: format timezone: timezone] ].
	self assert: (formatter value: #($- $T $: nil) value: 2 hours) = '2012-04-20T17:50:10+02:00'.
	self assert: (formatter value: #($- $T $: nil) value: -2 hours) = '2012-04-20T13:50:10-02:00'.
	self assert: (formatter value: #(nil $T nil nil) value: 1 hours) = '20120420T165010+0100'.
	self assert: (formatter value: #(nil $T nil nil) value: -1 hours) = '20120420T145010-0100'.
! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 11/21/2016 11:08'!
testToday
	self assert: ZTimestamp today secondsSinceMidnight isZero! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:03'!
testUnixTime
	| timestamp billenium |
	timestamp := ZTimestamp fromUnixTime: 1000000000.
	billenium := self unixTimeBillenium.
	self assert: timestamp = billenium.
	self assert: billenium asUnixTime = 1000000000! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 4/4/2014 07:36'!
testWithNowDo
	(Time respondsTo: #microsecondClockValue)
		ifFalse: [ self skip ].
	ZTimestamp withNowDo: [ :timestamp :microseconds |
		self assertInvariants: timestamp.
		self assert: (microseconds between: 0 and: 1e6 - 1) ]! !

!ZTimestampTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 6/12/2013 16:15'!
testZTimestampFormatting
	| timestamp |
	timestamp := self lunarLanding.
	self 
		assert: (timestamp format: '2001-02-03T16:05:06Z') 
		equals: '1969-07-20T20:17:40Z'.
	self 
		assert: (String streamContents: [ :out | timestamp format: '01/02/03 16:05:06' on: out ]) 
		equals: '69/07/20 20:17:40'.
! !

!ZTimestampTests methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/22/2012 11:02'!
unixTimeBillenium
	"1 billion seconds since the Unix Epoch"
	
	^ ZTimestamp fromString: '2001-09-09T01:46:40Z'.! !

TestCase subclass: #ZTimezoneTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 22:40'!
assertIsTimezoneEuropeBrussels: timezone
	| winter summer |
	winter := ZTimestamp year: 2014 month: 01 day: 02.
	self assert: (timezone offsetForTimestamp: winter) equals: 1 hour.
	self assert: (timezone subzoneForTimestamp: winter) abbreviation equals: #CET.
	self deny: (timezone subzoneForTimestamp: winter) isDaylightSaving.
	summer := ZTimestamp year: 2014 month: 08 day: 02.
	self assert: (timezone offsetForTimestamp: summer) equals: 2 hours.
	self assert: (timezone subzoneForTimestamp: summer) abbreviation equals: #CEST.
	self assert: (timezone subzoneForTimestamp: summer) isDaylightSaving! !

!ZTimezoneTests methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/11/2014 20:23'!
europeBrusselsContents
	"'/usr/share/zoneinfo/Europe/Brussels' asFileReference binaryReadStreamDo: [ :in | in upToEnd ]."
	
	^ #[84 90 105 102 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 10 0 0 0 10 0 0 0 0 0 0 0 185 0 0 0 10 0 0 0 18 152 68 73 128 155 12 37 112 155 213 218 240 156 217 174 144 157 164 181 144 158 185 144 144 159 132 151 144 159 206 248 48 160 96 165 240 161 126 187 112 162 46 18 240 163 122 76 240 164 53 129 240 165 94 35 112 166 37 53 240 167 39 155 240 168 42 1 240 169 7 125 240 169 238 52 112 170 231 95 240 171 215 80 240 172 199 65 240 173 201 167 240 174 167 35 240 175 160 79 112 176 135 5 240 177 137 107 240 178 112 76 160 179 114 178 160 180 80 46 160 181 73 90 32 182 48 16 160 183 50 118 160 184 15 242 160 184 255 227 160 185 239 212 160 186 214 139 32 187 216 241 32 188 200 226 32 189 184 211 32 190 159 137 160 191 152 181 32 192 155 27 32 193 120 151 32 194 104 136 32 195 88 121 32 196 63 47 160 197 56 91 32 198 58 193 32 199 88 214 160 199 218 9 160 200 74 25 32 204 231 75 16 205 169 23 144 206 162 67 16 207 146 52 16 208 91 191 96 208 110 94 144 209 114 22 16 210 78 64 144 211 145 64 16 212 75 35 144 13 42 253 112 13 164 99 144 14 139 26 16 15 132 69 144 16 116 54 144 17 100 39 144 18 84 24 144 19 77 68 16 20 51 250 144 21 35 235 144 22 19 220 144 23 3 205 144 23 243 190 144 24 227 175 144 25 211 160 144 26 195 145 144 27 188 189 16 28 172 174 16 29 156 159 16 30 140 144 16 31 124 129 16 32 108 114 16 33 92 99 16 34 76 84 16 35 60 69 16 36 44 54 16 37 28 39 16 38 12 24 16 39 5 67 144 39 245 52 144 40 229 37 144 41 213 22 144 42 197 7 144 43 180 248 144 44 164 233 144 45 148 218 144 46 132 203 144 47 116 188 144 48 100 173 144 49 93 217 16 50 114 180 16 51 61 187 16 52 82 150 16 53 29 157 16 54 50 120 16 54 253 127 16 56 27 148 144 56 221 97 16 57 251 118 144 58 189 67 16 59 219 88 144 60 166 95 144 61 187 58 144 62 134 65 144 63 155 28 144 64 102 35 144 65 132 57 16 66 70 5 144 67 100 27 16 68 37 231 144 69 67 253 16 70 5 201 144 71 35 223 16 71 238 230 16 73 3 193 16 73 206 200 16 74 227 163 16 75 174 170 16 76 204 191 144 77 142 140 16 78 172 161 144 79 110 110 16 80 140 131 144 81 87 138 144 82 108 101 144 83 55 108 144 84 76 71 144 85 23 78 144 86 44 41 144 86 247 48 144 88 21 70 16 88 215 18 144 89 245 40 16 90 182 244 144 91 213 10 16 92 160 17 16 93 180 236 16 94 127 243 16 95 148 206 16 96 95 213 16 97 125 234 144 98 63 183 16 99 93 204 144 100 31 153 16 101 61 174 144 102 8 181 144 103 29 144 144 103 232 151 144 104 253 114 144 105 200 121 144 106 221 84 144 107 168 91 144 108 198 113 16 109 136 61 144 110 166 83 16 111 104 31 144 112 134 53 16 113 81 60 16 114 102 23 16 115 49 30 16 116 69 249 16 117 17 0 16 118 47 21 144 118 240 226 16 120 14 247 144 120 208 196 16 121 238 217 144 122 176 166 16 123 206 187 144 124 153 194 144 125 174 157 144 126 121 164 144 127 142 127 144 1 4 1 2 3 2 3 7 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 6 5 2 3 2 3 2 4 3 2 3 2 3 1 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 8 9 0 0 0 0 0 0 0 0 14 16 0 4 0 0 28 32 1 8 0 0 14 16 0 4 0 0 28 32 1 8 0 0 14 16 1 13 0 0 0 0 0 0 0 0 0 0 0 0 0 0 28 32 1 8 0 0 14 16 0 4 87 69 84 0 67 69 84 0 67 69 83 84 0 87 69 83 84 0 0 0 1 1 0 1 1 1 1 1 0 0 0 0 0 0 0 1 1 1]! !

!ZTimezoneTests methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/11/2014 20:35'!
runningOnWindows
	"Pharo 3+"
	[ ^ OSPlatform current isWindows ] on: MessageNotUnderstood do: [ ].
	"Pharo 2-"
	[ ^ OSPlatform isWin32 ] on: MessageNotUnderstood do: [ ].
	"Give up"
	^ false! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/16/2014 18:45'!
testDaylightSavingTimeTransition
	"Daylight Saving Time (DST) creates a discontinuity 
	in the local wall time when the transition occurs.
	For timezone Europe/Brussels, on March 30, 2014, at 01:00 GMT,
	the offset goes from 1 hour (CET) to 2 hours (CEST), which means that
	the local wall time jumps from 02:00 to 03:00 on that day.
	Local time within that interval is ill defined."

	| timezone |
	timezone := ZTimezone id: 'Europe/Brussels'.
	"1 second before the transition"
	self
		assert: (timezone gmtToLocal: (ZTimestamp fromString: '2014-03-30T00:59:59Z'))
		equals: (ZTimestamp fromString: '2014-03-30T01:59:59Z').
	"a second later the transition occurs, jumping 1 hour forward"
	self
		assert: (timezone gmtToLocal: (ZTimestamp fromString: '2014-03-30T01:00:00Z'))
		equals: (ZTimestamp fromString: '2014-03-30T03:00:00Z').
	"reverse conversion just before the transition"
	self
		assert: (timezone localToGmt: (ZTimestamp fromString: '2014-03-30T01:59:59Z'))
		equals: (ZTimestamp fromString: '2014-03-30T00:59:59Z').	
	"reverse conversion a second later, after jumping 1 hour forward in local wall time"
	self
		assert: (timezone localToGmt: (ZTimestamp fromString: '2014-03-30T03:00:00Z'))
		equals: (ZTimestamp fromString: '2014-03-30T01:00:00Z').	
	"for an ill defined local wall time inside the jump, the offset is still 1 hour"
	self
		assert: (timezone localToGmt: (ZTimestamp fromString: '2014-03-30T02:30:00Z'))
		equals: (ZTimestamp fromString: '2014-03-30T01:30:00Z').
	"but converting that GMT timestamp gives a different local wall time"
	self
		assert: (timezone gmtToLocal: (ZTimestamp fromString: '2014-03-30T01:30:00Z'))
		equals: (ZTimestamp fromString: '2014-03-30T03:30:00Z').

! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 22:41'!
testEuropeBrussels
	| timezone |
	timezone := ZTimezone id: #'Europe/Brussels'.
	self assert: timezone id equals: #'Europe/Brussels'.
	self assertIsTimezoneEuropeBrussels: timezone! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 20:42'!
testGMT
	| timezone subTimezone |
	timezone := ZTimezone gmt.
	self assert: timezone id equals: #GMT.
	subTimezone := timezone subzoneForTimestamp: ZTimestamp now.
	self assert: subTimezone abbreviation equals: #GMT.
	self assert: subTimezone offset equals: Duration zero.
	self deny: subTimezone isDaylightSaving.
	self assert: (ZTimezone id: #GMT) equals: timezone! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/16/2014 16:25'!
testGmtToLocal
	| timezone |
	timezone := ZTimezone id: 'Europe/Brussels'.
	self 
		assert: (timezone gmtToLocal: (ZTimestamp fromString: '2014-01-15T12:00:00Z'))
		equals: (ZTimestamp fromString: '2014-01-15T13:00:00Z').
	self 
		assert: (timezone gmtToLocal: (ZTimestamp fromString: '2014-08-15T12:00:00Z'))
		equals: (ZTimestamp fromString: '2014-08-15T14:00:00Z').! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 20:40'!
testIdentifiers
	self runningOnWindows ifTrue: [ self skip ].
	self deny: ZTimezone timezoneIdentifiers isEmpty.
	self assert: (ZTimezone timezoneIdentifiers includes: #'Europe/Brussels')! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/16/2014 16:26'!
testLocalToGmt
	| timezone |
	timezone := ZTimezone id: 'Europe/Brussels'.
	self 
		assert: (timezone localToGmt: (ZTimestamp fromString: '2014-01-15T12:00:00Z'))
		equals: (ZTimestamp fromString: '2014-01-15T11:00:00Z').
	self 
		assert: (timezone localToGmt: (ZTimestamp fromString: '2014-08-15T12:00:00Z'))
		equals: (ZTimestamp fromString: '2014-08-15T10:00:00Z').! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 22:41'!
testParsing
	| timezone |
	timezone := ZTimezone fromStream: self europeBrusselsContents readStream.
	self assert: timezone transitions size equals: 185.
	self assert: timezone subzones size equals: 10.
	self assertIsTimezoneEuropeBrussels: timezone! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/12/2014 16:20'!
testReadAll
	| timezones timezone |
	self runningOnWindows ifTrue: [ self skip ].
	timezones := ZTimezone timezoneIdentifiers 
		collect: [ :each | ZTimezone readTimezoneWithId: each ].
	self deny: timezones isEmpty.
	timezone := timezones 
		detect: [ :each | each id = #'Europe/Brussels' ]
		ifNone: [ self fail ].
	self assertIsTimezoneEuropeBrussels: timezone! !

!ZTimezoneTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 3/11/2014 20:43'!
testZoneInfoLocation
	self runningOnWindows ifTrue: [ self skip ].
	self assert: ZTimezone zoneInfoLocation exists! !

!DateAndTime methodsFor: '*ztimestamp' stamp: 'SvenVanCaekenberghe 3/29/2014 00:43'!
asZTimestamp
	"Convert the receiver to a ZTimestamp.
	Note that we are implicitly converting to UTC by ignoring the offset.
	Note also that nano seconds are dropped."

	^ ZTimestamp new
		jdn: julianDayNumber 
		secs: seconds! !

!DateAndTime methodsFor: '*ztimestamp' stamp: 'SvenVanCaekenberghe 3/28/2014 17:08'!
subtractFromTimestamp: timestamp
	"Part of the double dispatch implementation of ZTimestamp>>#-"
	
	^ timestamp subtractTimestamp: self asZTimestamp! !

Object subclass: #ZSubTimezone
	instanceVariableNames: 'abbreviation offset isDaylightSaving'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZSubTimezone commentStamp: 'SvenVanCaekenberghe 3/12/2014 13:12' prior: 0!
I am ZSubTimezone.
I represent a specific offset duration to be added to UTC.
I am identified by  an abbreviation.
I know whether daylight saving time is in use or not.

See ZTimezone!

!ZSubTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/11/2014 00:08'!
abbreviation: string offset: integer dst: boolean
	^ self new
		abbreviation: string;
		offset: integer;
		isDaylightSaving: boolean;
		yourself! !

!ZSubTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/11/2014 13:23'!
gmt
	^ self abbreviation: #GMT offset: 0 dst: false! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 00:07'!
abbreviation
	^ abbreviation! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:24'!
abbreviation: string
	abbreviation := string asSymbol! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 00:07'!
isDaylightSaving
	^ isDaylightSaving! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:24'!
isDaylightSaving: boolean
	isDaylightSaving := boolean! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 00:07'!
offset
	^ offset! !

!ZSubTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:25'!
offset: integer
	offset := integer asDuration! !

!ZSubTimezone methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 3/11/2014 00:12'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream << abbreviation; space; print: offset.
	isDaylightSaving ifTrue: [ stream << ' DST' ]. 
	stream nextPut: $)! !

Object subclass: #ZTimestampFormat
	instanceVariableNames: 'specification formatter parser constructor weekdayNames monthNames timezone'
	classVariableNames: 'FormatKeys Formats'
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampFormat commentStamp: 'SvenVanCaekenberghe 3/12/2014 22:45' prior: 0!
I am ZTimestampFormat, an implementation of a textual representation for a timestamp, date or time that can be used for formatting or parsing.

You instanciate me by specifying the textual format by example, based on a #reference timetamp. 
Each component of the example representation is numbered from largest to smallest unit:
	1=year
	2=month
	3=dayInMonth
	4=hour (16 in 24 hour format)
	5=minute
	6=second
as in the ISO representation: 
	2001-02-03T16:05:06Z which is a Saterday. 
Example format strings can be found in my class accessing protocol or in the unit tests.

To specifiy a format, you write the reference date so that it matches the representation that you want.

	(ZTimestampFormat fromString: 'SAT, FEB 03 2001 (16:05:06)')
		format: ZTimestamp now.

I can be used for unabiguous, stricter parsing as well.

	(ZTimestampFormat fromString: '02/03/01 (16:05:06)')
		parse: '10/10/10 (12:01:01)'.
		
The list of possible keys and their interpretation #formatSpecifications.
I can translate month and weekday names to 4 different languages, English, French, German and Dutch.
I can optionally use a timezone to convert UTC/GMT/Zulu timestamps to local time.!

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 4/15/2014 14:52'!
abbreviatedTimeZones
	"Map TimeZone abbreviations to their offsets"
	"Order matters for lookup by offset !!"
	"Initially created then edited by hand using"
	"TimeZone timeZones collect: [ :each | { each abbreviation. each offset hours } ]"

	^ #(
		#('UTC' 0) 
		#('GMT' 0) 
		#('CET' 1)
		#('BST' 1) 
		#('SAST' 2) 
		#('PST' -8) 
		#('PDT' -7) 
		#('MST' -7) 
		#('MDT' -6) 
		#('CST' -6) 
		#('CDT' -5) 
		#('EST' -5) 
		#('EDT' -4)
		#('CEST' 2) 
		)! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 16:48'!
ansiC
	^ self fromString: 'Sat Feb _3 16:05:06 2001'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/3/2012 19:01'!
formatKeys
	^ FormatKeys ! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 6/10/2015 16:13'!
formatSpecifications
	^ #(
		'2001' 		yearFourDigits
		'01' 			yearTwoDigits
		'02001'		yearFull
		'02'			monthTwoDigits
		'2'			month
		'February'	monthNameCapitalized
		'february'	monthNameLowercased
		'FEBRUARY'	monthNameUppercased
		'Feb'			monthNameAbbreviatedCapitalized
		'feb'			monthNameAbbreviatedLowercased
		'FEB'			monthNameAbbreviatedUppercased
		'3'			day
		'_3'			dayTwoDigitsSpacePadded
		'03'			dayTwoDigits
		'4'			hour12
		'04'			hour12TwoDigits
		'16'			hour24TwoDigits
		'PM'			daypartUppercased
		'pm'			daypartLowercased
		'5'			minute
		'05'			minuteTwoDigits
		'6'			second
		'06'			secondTwoDigits
		'7'			nanosecond
		'Z'			timeZoneZ
		'+00:00'	timeZone
		'UTC'			timeZoneAbbreviated
		'_UTC'		timeZoneAbbreviatedLong
		'Saturday'	weekdayNameCapitalized
		'saturday'	weekdayNameLowercased
		'SATURDAY'	weekdayNameUppercased
		'Sat'			weekdayNameAbbreviatedCapitalized
		'sat'			weekdayNameAbbreviatedLowercased
		'SAT'			weekdayNameAbbreviatedUppercased )! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/3/2012 19:01'!
formats
	^ Formats! !

!ZTimestampFormat class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 10/29/2012 14:40'!
fromString: example
	^ self new
		specification: example;
		yourself
	! !

!ZTimestampFormat class methodsFor: 'class initialization' stamp: 'SvenVanCaekenberghe 4/15/2014 14:55'!
initialize
	"Initialize this class, to be redone every time #formatSpecifications changes."
	
	"self initialize."
	
	Formats := Dictionary newFromPairs: self formatSpecifications.
	FormatKeys := Formats keys sorted: [ :x :y | x size >= y size ]! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/12/2014 23:23'!
iso
	^ self fromString: '2001-02-03T16:05:06Z'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/13/2014 15:52'!
isoCompact
	^ self fromString: '20010203T160506'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/13/2014 15:52'!
isoTZ
	^ self fromString: '2001-02-03T16:05:06+00:00'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 20:43'!
kitchen
	^ self fromString: '3:04PM'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 20:38'!
reference12
	"The ZTimestamp instance used as a reference for specifying a format by example.
	The possible components are numbered from largest to smallest unit."

	^ ZTimestamp year: 2001 month: 2 day: 3 hour: 4 minute: 5 second: 6! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 20:41'!
reference24
	"The ZTimestamp instance used as a reference for specifying a format by example.
	The possible components are numbered from largest to smallest unit."

	^ ZTimestamp year: 2001 month: 2 day: 3 hour: 16 minute: 5 second: 6! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 16:54'!
rfc3339
	^ self fromString: '2001-02-03T16:05:06Z'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 16:49'!
rfc822
	^ self fromString: '03 Feb 01 16:05 UTC'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/1/2014 15:36'!
rfc850
	^ self fromString: 'Saturday, 03-Feb-01 16:05:06 UTC'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 16:49'!
unixDate
	^ self fromString: 'Sat Feb _3 16:05:06 UTC 2001'! !

!ZTimestampFormat class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/1/2014 15:36'!
verbose
	^ self fromString: 'Saturday, 03-February-2001 04:05:06 PM (+00:00)'! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/12/2014 14:44'!
abbreviationForTimeZoneOffset: offset
	"(TimeZone offset: offset) abbreviation"

	self class abbreviatedTimeZones
		do: [ :each | 
			each second = offset hours
				ifTrue: [ ^ each first ] ].
	self error: 'Cannot find time zone for offset ' , offset asString! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 23:36'!
actualOffsetForTimestamp: timestamp
	^ self timezone value
		ifNil: [ timestamp offset ]
		ifNotNil: [ :actualTimezone | actualTimezone offsetForTimestamp: timestamp ]! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/3/2012 19:11'!
compileSpecification
	| offset elements |
	elements := Array streamContents: [ :stream |
		offset := 1.
		[ offset <= specification size ] whileTrue: [ 
			(self nextFormatKeyStartingAt: offset)
				ifNotNil: [ :formatKey |
					stream nextPut: (self class formats at: formatKey).
					offset := offset + formatKey size ]
				ifNil: [ 
					stream nextPut: (specification at: offset).
					offset := offset + 1 ] ] ].
	formatter := elements collect: [ :each | 
			each isCharacter ifTrue: [ each ] ifFalse: [ (#format:, each, #On:) asSymbol ] ].
	parser := elements collect: [ :each | 
			each isCharacter ifTrue: [ each ] ifFalse: [ (#parse, each capitalized, #From:) asSymbol ] ] ! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:40'!
constructor
	^ constructor! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:27'!
constructor: anObject
	constructor := anObject! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:31'!
createDate
	self constructor: [ :elements | 
		Date 
			year: (elements at: #year ifAbsent: [ 0 ]) 
			month: (elements at: #month ifAbsent: [ 0 ])  
			day: (elements at: #day ifAbsent: [ 0 ]) ]! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 6/10/2015 16:22'!
createDateAndTime
	self constructor: [ :elements | 
		| timestamp |
		timestamp := DateAndTime 
			year: (elements at: #year ifAbsent: [ 0 ]) 
			month: (elements at: #month ifAbsent: [ 0 ])  
			day: (elements at: #day ifAbsent: [ 0 ])  
			hour: (elements at: #hour ifAbsent: [ 0 ]) 
			minute: (elements at: #minute ifAbsent: [ 0 ])  
			second: (elements at: #second ifAbsent: [ 0 ]) 
			nanoSecond: (elements at: #nanosecond ifAbsent: [ 0 ])
			offset: (elements at: #offset ifAbsent: [ Duration zero ]).
		(elements includesKey: #offset)
			ifTrue: [ timestamp ]
			ifFalse: [ self localToGmt: timestamp ] ]! !

!ZTimestampFormat methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:19'!
createInstanceFrom: elements
	elements 
		at: #daypart 
		ifPresent: [ :daypart | | hour |
			hour :=  elements at: #hour.
			(daypart = #pm and: [ hour < 12 ])
				ifTrue: [ elements at: #hour put: hour + 12 ].
			(daypart = #am and: [ hour = 12 ])
				ifTrue: [ elements at: #hour put: 0 ] ].
	^ self constructor value: elements! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:31'!
createTime
	self constructor: [ :elements | 
		Time 
			hour: (elements at: #hour ifAbsent: [ 0 ]) 
			minute: (elements at: #minute ifAbsent: [ 0 ])  
			second: (elements at: #second ifAbsent: [ 0 ]) ]! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 3/13/2014 15:19'!
createZTimestamp
	self constructor: [ :elements | 
		| timestamp |
		timestamp := ZTimestamp 
			year: (elements at: #year ifAbsent: [ 0 ]) 
			month: (elements at: #month ifAbsent: [ 0 ])  
			day: (elements at: #day ifAbsent: [ 0 ])  
			hour: (elements at: #hour ifAbsent: [ 0 ]) 
			minute: (elements at: #minute ifAbsent: [ 0 ])  
			second: (elements at: #second ifAbsent: [ 0 ]) 
			offset: (elements at: #offset ifAbsent: [ 0 ]).
		(elements includesKey: #offset)
			ifTrue: [ timestamp ]
			ifFalse: [ self localToGmt: timestamp ] ]! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:34'!
dutch
	self monthNames: #( 
		'Januari' 'Februari' 'Maart' 'April' 'Mei' 'Juni' 
		'Juli' 'Augustus' 'September' 'Oktober' 'November' 'December' ).
	self weekdayNames: #( 
		'Zondag' 'Maandag' 'Dinsdag' 'Woensdag' 'Donderdag' 'Vrijdag' 'Zaterdag' )! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 2/1/2014 15:36'!
english
	self monthNames: #( 
		January February March April May June 
		July August September October November December).
	self weekdayNames: #( 
		Sunday Monday Tuesday Wednesday Thursday Friday Saturday )! !

!ZTimestampFormat methodsFor: 'formatting' stamp: 'SvenVanCaekenberghe 6/12/2013 16:00'!
format: object
	^ String streamContents: [ :stream |
			self format: object on: stream ]! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:26'!
format: object dayOn: stream
	object dayOfMonth printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:26'!
format: object dayTwoDigitsOn: stream
	object dayOfMonth printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:18'!
format: object dayTwoDigitsSpacePaddedOn: stream
	object dayOfMonth < 10
		ifTrue: [ stream space ].
	object dayOfMonth printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:16'!
format: object daypartLowercasedOn: stream
	stream << (object hour < 12 ifTrue: [ #am ] ifFalse: [ #pm ])! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:16'!
format: object daypartUppercasedOn: stream
	stream << (object hour < 12 ifTrue: [ #AM ] ifFalse: [ #PM ])! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:13'!
format: object hour12On: stream
	| hour |
	hour := object hour.
	(hour > 12
		ifTrue: [ hour - 12 ]
		ifFalse: [ 
			hour < 1
				ifTrue: [ 12 ]
				ifFalse: [ hour ] ]) printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:14'!
format: object hour12TwoDigitsOn: stream
	| hour |
	hour := object hour.
	(hour > 12
		ifTrue: [ hour - 12 ]
		ifFalse: [ 
			hour < 1
				ifTrue: [ 12 ]
				ifFalse: [ hour ] ]) printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:15'!
format: object hour24TwoDigitsOn: stream
	object hour printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:10'!
format: object minuteOn: stream
	object minute printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:10'!
format: object minuteTwoDigitsOn: stream
	object minute printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameAbbreviatedCapitalizedOn: stream
	stream << ((self monthNames at: object monthIndex) copyFrom: 1 to: 3)! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameAbbreviatedLowercasedOn: stream
	stream << ((self monthNames at: object monthIndex) copyFrom: 1 to: 3) asLowercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameAbbreviatedUppercasedOn: stream
	stream << ((self monthNames at: object monthIndex) copyFrom: 1 to: 3) asUppercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameCapitalizedOn: stream
	stream << (self monthNames at: object monthIndex)! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameLowercasedOn: stream
	stream << (self monthNames at: object monthIndex) asLowercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:13'!
format: object monthNameUppercasedOn: stream
	stream << (self monthNames at: object monthIndex) asUppercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 2/1/2014 13:14'!
format: object monthOn: stream
	object monthIndex printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 12/19/2012 13:27'!
format: object monthTwoDigitsOn: stream
	object monthIndex printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 6/10/2015 16:27'!
format: object nanosecondOn: stream
	| n len |
	n := object nanoSecond.
	n ~=0
		ifTrue: [  
			len := 9.
			[ n \\ 10 = 0 ] whileTrue: [ 
				n := n / 10.
				len := len - 1 ] ] 
		ifFalse: [ len := 1 ].
	n printOn: stream base: 10 length: len padded: true! !

!ZTimestampFormat methodsFor: 'formatting' stamp: 'SvenVanCaekenberghe 3/12/2014 23:21'!
format: object on: stream
	| local |
	local := self gmtToLocal: object.
	formatter do: [ :each |
		each isCharacter 
			ifTrue: [ stream nextPut: each ] 
			ifFalse: [ self perform: each with: local with: stream ] ]! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:06'!
format: object secondOn: stream
	object second printOn: stream! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 20:06'!
format: object secondTwoDigitsOn: stream
	object second printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 4/15/2014 14:43'!
format: object timeZoneAbbreviatedLongOn: stream
	| offset |
	offset := self actualOffsetForTimestamp: object.
	offset isZero
		ifTrue: [ stream << #UTC ]
		ifFalse: [ stream << (self abbreviationForTimeZoneOffset: offset) ]! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 3/12/2014 23:38'!
format: object timeZoneAbbreviatedOn: stream
	| offset |
	offset := self actualOffsetForTimestamp: object.
	offset isZero
		ifTrue: [ stream << #UTC ]
		ifFalse: [ stream << (self abbreviationForTimeZoneOffset: offset) ]! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 3/12/2014 23:39'!
format: object timeZoneOn: stream
	| offset |
	offset := self actualOffsetForTimestamp: object.
	stream nextPut: (offset positive ifTrue: [ $+ ] ifFalse: [ $- ]).
	offset hours abs
		printOn: stream base: 10 length: 2 padded: true.
	stream nextPut: $:.
	offset minutes
		printOn: stream base: 10 length: 2 padded: true ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 3/12/2014 23:40'!
format: object timeZoneZOn: stream
	| offset |
	offset := self actualOffsetForTimestamp: object.
	offset isZero
		ifTrue: [ stream nextPut: $Z ]
		ifFalse: [ 
			stream nextPut: (offset positive ifTrue: [ $+ ] ifFalse: [ $- ]).
			offset hours abs
				printOn: stream base: 10 length: 2 padded: true.
			stream nextPut: $:.
			offset minutes
				printOn: stream base: 10 length: 2 padded: true ]! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:52'!
format: object weekdayNameAbbreviatedCapitalizedOn: stream
	stream << ((self weekdayNames at: object dayOfWeek) copyFrom: 1 to: 3)! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:52'!
format: object weekdayNameAbbreviatedLowercasedOn: stream
	stream << ((self weekdayNames at: object dayOfWeek) copyFrom: 1 to: 3) asLowercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:52'!
format: object weekdayNameAbbreviatedUppercasedOn: stream
	stream << ((self weekdayNames at: object dayOfWeek) copyFrom: 1 to: 3) asUppercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:51'!
format: object weekdayNameCapitalizedOn: stream
	stream << (self weekdayNames at: object dayOfWeek)! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:51'!
format: object weekdayNameLowercasedOn: stream
	stream << (self weekdayNames at: object dayOfWeek) asLowercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:51'!
format: object weekdayNameUppercasedOn: stream
	stream << (self weekdayNames at: object dayOfWeek) asUppercase ! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 3/13/2014 15:56'!
format: object yearFourDigitsOn: stream
	object year printOn: stream base: 10 length: 4 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:43'!
format: object yearFullOn: stream
	object year printOn: stream base: 10 length: 4 padded: true! !

!ZTimestampFormat methodsFor: 'formatters' stamp: 'SvenVanCaekenberghe 11/3/2012 19:43'!
format: object yearTwoDigitsOn: stream
	object year \\ 100 printOn: stream base: 10 length: 2 padded: true! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/3/2012 19:05'!
formatter
	^ formatter! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:35'!
french
	self monthNames: #( 
		'Janvier' 'Février' 'Mars' 'Avril' 'Mai' 'Juin' 
		'Juillet' 'Août' 'Septembre' 'Octobre' 'Novembre' 'Décembre' ).
	self weekdayNames: #( 
		'Dimanche' 'Lundi' 'Mardi' 'Mecredi' 'Jeudi' 'Vendredi' 'Samedi' )! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:35'!
german
	self monthNames: #( 
		'Januar' 'Februar' 'März' 'April' 'Mai' 'Juni' 
		'Juli' 'August' 'September' 'Oktober' 'November' 'Dezember' ).
	self weekdayNames: #( 
		'Sonntag' 'Montag' 'Dienstag' 'Mittwoch' 'Donnerstag' 'Freitag' 'Samstag' )! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/15/2014 13:17'!
gmtToLocal: timestamp
	^ self timezone value
		ifNil: [ timestamp ]
		ifNotNil: [ :actualTimezone |actualTimezone gmtToLocal: timestamp ]! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/1/2012 10:40'!
initialize 
	super initialize.
	self english; createZTimestamp! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/15/2014 13:17'!
localToGmt: timestamp
	^ self timezone value 
		ifNil: [ timestamp ] 
		ifNotNil: [ :actualTimezone | actualTimezone localToGmt: timestamp ]! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:41'!
monthNames
	^ monthNames! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 13:52'!
monthNames: anObject
	monthNames := anObject! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/3/2012 20:12'!
nextFormatKeyStartingAt: offset
	^ self class formatKeys
		detect: [ :each | self string: specification contains: each startingAt: offset ]
		ifNone: [ nil ]! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/12/2014 14:44'!
offsetForTimeZoneAbbreviation: abbreviation
	"(TimeZone abbreviated: abbreviation) offset"

	self class abbreviatedTimeZones
		do: [ :each | 
			each first = abbreviation
				ifTrue: [ ^ each second hours ] ].
	self error: 'Cannot find time zone for abbreviation ' , abbreviation asString! !

!ZTimestampFormat methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 10/30/2012 10:28'!
parse: string
	^ self readFrom: string readStream! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:41'!
parseDayFrom: stream
	^ #day -> (self readInteger: 1 atLeastFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:42'!
parseDayTwoDigitsFrom: stream
	^ #day -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:42'!
parseDayTwoDigitsSpacePaddedFrom: stream
	^ #day -> (self readInteger: 2 optionalSpaceFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:40'!
parseDaypartLowercasedFrom: stream
	| part |
	part := stream next: 2.
	self assert: (#(#am #pm) includes: part).
	^ #daypart -> part asSymbol! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:39'!
parseDaypartUppercasedFrom: stream
	| part |
	part := stream next: 2.
	self assert: (#(#AM #PM) includes: part).
	^ #daypart -> part asLowercase asSymbol! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:38'!
parseHour12From: stream
	^ #hour -> (self readInteger: 1 atLeastFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:38'!
parseHour12TwoDigitsFrom: stream
	^ #hour -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:38'!
parseHour24TwoDigitsFrom: stream
	^ #hour -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:36'!
parseMinuteFrom: stream
	^ #minute -> (self readInteger: 1 atLeastFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:30'!
parseMinuteTwoDigitsFrom: stream
	^ #minute -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:36'!
parseMonthFrom: stream
	^ #month -> (self readInteger: 1 atLeastFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameAbbreviatedCapitalizedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self monthNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) capitalized = abbreviation.
	^ #month -> (self monthNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameAbbreviatedLowercasedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self monthNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) asLowercase = abbreviation.
	^ #month -> (self monthNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameAbbreviatedUppercasedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self monthNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) asUppercase = abbreviation.
	^ #month -> (self monthNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameCapitalizedFrom: stream
	| match |
	match := self readAny: self monthNames sorted from: stream.
	self assert: match notNil.
	^ #month -> (self monthNames indexOf: match)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameLowercasedFrom: stream
	| match |
	match := self 
		readAny: (self monthNames collect: [ :each | each asLowercase ]) sorted 
		from: stream.
	self assert: match notNil.
	^ #month -> (self monthNames indexOf: match capitalized)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseMonthNameUppercasedFrom: stream
	| match |
	match := self 
		readAny: (self monthNames collect: [ :each | each asUppercase ]) sorted 
		from: stream.
	self assert: match notNil.
	^ #month -> (self monthNames indexOf: match asLowercase capitalized)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:27'!
parseMonthTwoDigitsFrom: stream
	^ #month -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 7/3/2015 13:54'!
parseNanosecondFrom: stream
	| fraction power |
	fraction := 0.
	power := 1.
	[ stream atEnd not and: [ stream peek isDigit ] ]
		whileTrue: [ 
			fraction := fraction * 10 + stream next digitValue.
			power := power * 10 ].
	^ #nanosecond -> (fraction / power * 1e9)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:35'!
parseSecondFrom: stream
	^ #second -> (self readInteger: 1 atLeastFrom: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:30'!
parseSecondTwoDigitsFrom: stream
	^ #second -> (self readInteger: 2 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 2/12/2014 14:30'!
parseTimeZoneAbbreviatedFrom: stream
	^ #offset -> (self offsetForTimeZoneAbbreviation: (stream next: 3))! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 4/15/2014 14:47'!
parseTimeZoneAbbreviatedLongFrom: stream
	| offset |
	offset := String streamContents: [ :out |
		[ stream atEnd not and: [ 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' includes: stream peek ] ]
			whileTrue: [ out nextPut: stream next ] ].
	^ #offset -> (self offsetForTimeZoneAbbreviation: offset)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:33'!
parseTimeZoneFrom: stream
	^ #offset -> (ZTimestamp readTimezoneOffsetFrom: stream) seconds! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:32'!
parseTimeZoneZFrom: stream
	^ #offset -> (ZTimestamp readTimezoneOffsetFrom: stream) seconds! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseWeekdayNameAbbreviatedCapitalizedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self weekdayNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) capitalized = abbreviation.
	^ #weekday -> (self weekdayNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseWeekdayNameAbbreviatedLowercasedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self weekdayNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) asLowercase = abbreviation.
	^ #weekday -> (self weekdayNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:28'!
parseWeekdayNameAbbreviatedUppercasedFrom: stream
	| abbreviation fullName target |
	abbreviation := stream next: 3.
	target := abbreviation asLowercase capitalized.
	fullName := self weekdayNames detect: [ :each | 
		self string: each contains: target startingAt: 1 ].
	self assert: (fullName copyFrom: 1 to: 3) asUppercase = abbreviation.
	^ #weekday -> (self weekdayNames indexOf: fullName)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:29'!
parseWeekdayNameCapitalizedFrom: stream
	| match |
	match := self readAny: self weekdayNames sorted from: stream.
	self assert: match notNil.
	^ #weekday -> (self weekdayNames indexOf: match)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:29'!
parseWeekdayNameLowercasedFrom: stream
	| match |
	match := self 
		readAny: (self weekdayNames collect: [ :each | each asLowercase ]) sorted 
		from: stream.
	self assert: match notNil.
	^ #weekday -> (self weekdayNames indexOf: match capitalized)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 20:29'!
parseWeekdayNameUppercasedFrom: stream
	| match |
	match := self 
		readAny: (self weekdayNames collect: [ :each | each asUppercase ]) sorted 
		from: stream.
	self assert: match notNil.
	^ #weekday -> (self weekdayNames indexOf: match asLowercase capitalized)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 3/13/2014 16:02'!
parseYearFourDigitsFrom: stream
	^ #year -> (self readInteger: 4 from: stream)! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:31'!
parseYearFullFrom: stream
	| sign |
	sign := 1.
	stream peek = $-
		ifTrue: [ 
			stream next.
			sign := -1 ].
	^ #year -> (sign * (self readInteger: 4 atLeastFrom: stream))! !

!ZTimestampFormat methodsFor: 'parsers' stamp: 'SvenVanCaekenberghe 11/3/2012 19:31'!
parseYearTwoDigitsFrom: stream
	^ #year -> (2000 + (self readInteger: 2 from: stream))! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/3/2012 19:05'!
parser
	^ parser! !

!ZTimestampFormat methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:51'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream << specification.
	stream nextPut: $)! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 11/5/2012 20:06'!
readAny: strings from: stream
	"Read any of the sorted collection strings from stream, or nil"
	
	| candidate matchCount |
	candidate := 1.
	matchCount := 0.
	[
		(strings at: candidate) size = matchCount 
			ifTrue: [ ^ strings at: candidate ].
		(stream atEnd or: [ candidate > strings size ])
			ifTrue: [ ^ nil ].
		(((strings at: candidate) at: matchCount + 1) = stream peek)
			ifTrue: [
				stream next.
				matchCount := matchCount + 1 ]
			ifFalse: [
				candidate := candidate + 1.
				(candidate <= strings size
					and: [ (1 to: matchCount) allSatisfy: [ :each | 
								((strings at: candidate - 1) at: each) = ((strings at: candidate) at: each) ] ])
					ifFalse: [ ^ nil ] ]
	] repeat! !

!ZTimestampFormat methodsFor: 'parsing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:21'!
readFrom: stream
	| offset elements |
	elements := SmallDictionary new.
	parser do: [ :each |
		each isCharacter 
			ifTrue: [ self assert: stream next = each ]
			ifFalse: [ elements add: (self perform: each with: stream) ] ].
	^ self createInstanceFrom: elements! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 10/30/2012 14:31'!
readInteger: numberOfDigits atLeastFrom: stream
	"Parse and return a decimal number of at least numberOfDigits from stream.
	Continue reading until a non digit or the end of stream is reached.
	Fail if that is not possible"

	| integer |
	integer := 0.
	numberOfDigits timesRepeat: [ | char |
		char := stream next.
		self assert: ('0123456789' includes: char).
		integer := (integer * 10) + char digitValue ].
	[ stream atEnd not and: [ stream peek isDigit ] ] whileTrue: [
		integer := (integer * 10) + stream next digitValue ].
	^ integer! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 10/29/2012 21:39'!
readInteger: numberOfDigits from: stream
	"Parse and return a decimal number of numberOfDigits from stream.
	Fail if that is not possible"

	| integer |
	integer := 0.
	numberOfDigits timesRepeat: [ | char |
		char := stream next.
		self assert: ('0123456789' includes: char).
		integer := (integer * 10) + char digitValue ].
	^ integer! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 10/30/2012 14:33'!
readInteger: numberOfDigits optionalSpaceFrom: stream
	"Parse and return a decimal number of numberOfDigits from stream.
	There can be one optional space instead of a digit at the beginning.
	Fail if that is not possible"

	^ stream peek = $ 
		ifTrue: [ 
			stream next. 
			self readInteger: numberOfDigits - 1 from: stream ]
		ifFalse: [
			self readInteger: numberOfDigits from: stream ]! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/3/2012 20:46'!
specification
	^ specification ! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 11/3/2012 19:04'!
specification: string
	specification := string.
	self compileSpecification! !

!ZTimestampFormat methodsFor: 'private' stamp: 'SvenVanCaekenberghe 10/29/2012 15:37'!
string: string contains: subString startingAt: offset
	subString doWithIndex: [ :each :index |
		offset + index - 1 <= string size 
			ifFalse: [ ^ false ].
		(string at: offset + index - 1) = each
			ifFalse: [ ^ false ] ].
	^ true! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/12/2014 23:11'!
timezone
	^ timezone! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/12/2014 23:17'!
timezone: anObject
	"Set the optional timezone that I should use to convert UTC/GMT/Zulu timestamps 
	into local time representations and vice versa if none is specified.
	The timezone object will be sent #value before use.
	When the value is nil, no conversions will be done.
	The returned value will be sent #offsetForTimestamp: when used."
	
	timezone := anObject! !

!ZTimestampFormat methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 3/12/2014 23:12'!
useCurrentTimezone
	self timezone: ZCurrentTimezone! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 11/1/2012 10:41'!
weekdayNames
	^ weekdayNames! !

!ZTimestampFormat methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 10/29/2012 13:52'!
weekdayNames: anObject
	weekdayNames := anObject! !

Object subclass: #ZTimestampSNTPClient
	instanceVariableNames: 'socket serverAddress'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampSNTPClient commentStamp: '<historical>' prior: 0!
I am ZTimestampSNTPClient, I can access an NTP server to retrieve an external time stamp.

I can be used to verify the current clock.

	ZTimestampSNTPClient new clockDifference.
	
	ZTimestampSNTPClient new
		enforceClockDifference: 2 seconds
		ifFail: [ :delta | 
			self inform: ('Clock difference {1} > 2s' format: { delta }) ];
		close.
		
	ZTimestampSNTPClient logClockDifferenceLargerThan: 1 second. 

Normally this difference should be just one or two seconds. If not, make sure your computer's clock itself is properly synchronized.

Note that following the design of ZTimestamp we work in second precision.

See 
	http://en.wikipedia.org/wiki/Network_Time_Protocol
	http://tools.ietf.org/html/rfc2030!

ZTimestampSNTPClient subclass: #ZTimestampPreciseSNTPClient
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimestampPreciseSNTPClient commentStamp: '<historical>' prior: 0!
I am ZTimestampPreciseSNTPClient.
I am a ZTimestampSNTPClient.

I use DataAndTime with theoretical nanosecond precision.

I can be used to verify the current clock.

	ZTimestampPreciseSNTPClient new clockDifference.
	
	ZTimestampPreciseSNTPClient new
		enforceClockDifference: 2 seconds
		ifFail: [ :delta | 
			self inform: ('Clock difference {1} > 2s' format: { delta }) ];
		close. 
		
	ZTimestampPreciseSNTPClient logClockDifferenceLargerThan: 1 second.

Normally this difference should be just just fractions of a seconds. If not, make sure your computer's clock itself is properly synchronized.!

!ZTimestampPreciseSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:04'!
now
	^ DateAndTime now asUTC! !

!ZTimestampPreciseSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/26/2013 00:25'!
ntpEpoch
	^ DateAndTime year: 1900 month: 1 day: 1 hour: 0 minute: 0 second: 0 offset: Duration zero! !

!ZTimestampPreciseSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/26/2013 09:53'!
timestampFromSeconds: seconds fraction: fraction
	^ self ntpEpoch + seconds seconds + (fraction * (10 raisedToInteger: 9)) nanoSeconds ! !

!ZTimestampSNTPClient class methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 5/31/2013 09:35'!
logClockDifferenceLargerThan: duration
	"self logClockDifferenceLargerThan: 1 second"
	
	self new
		enforceClockDifference: duration
		ifFail: [ :delta |
			self crLog:  ('Warning: clock difference {1} > {2}' format: { delta. duration }) ];
		close! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/26/2013 09:44'!
clockDifference
	"Return a Duration expressing the difference between the remote and local clocks"
	
	| now |
	now := self now.
	^ self remoteClock - now! !

!ZTimestampSNTPClient methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 2/25/2013 14:54'!
close
	socket close! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/25/2013 17:04'!
enforceClockDifference: duration ifFail: block 
	"Check the difference between the remote and local clock,
	and execute block if it is larger than duration."
	
	| difference |
	difference := self clockDifference.
	difference > duration
		ifTrue: [ block cull: difference ]! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 20:19'!
fractionalSecondsAt: offset in: bytes
	^ (bytes copyFrom: offset to: offset + 3) asInteger / (2 raisedToInteger: 32)! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 20:18'!
fullSecondsAt: offset in: bytes
	^ (bytes copyFrom: offset to: offset + 3) asInteger! !

!ZTimestampSNTPClient methodsFor: 'initialize-release' stamp: 'SvenVanCaekenberghe 2/25/2013 15:28'!
initialize
	socket := Socket newUDP.
	serverAddress := self ntpServerAddress ! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 23:33'!
interpretResult: bytes t0: t0 t1: t1
	| t2 t3 |
	t2 := self timestampAt: self receiveOffset in: bytes.
	t3 := self timestampAt: self transmitOffset in: bytes.
	^ Dictionary newFromPairs: { 
		#timestamp. t3.
		#roundtripDelay. (t3 - t0) - (t2 - t1).
		#localClockOffset. ((t1 - t0) + (t2 - t3)) / 2.
		#datagram. bytes }! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/26/2013 09:43'!
localClockOffset
	"Return a Delay representing the local clock offiset 
	as computed using the NTP clock synchronization algorithm"

	| result |
	result := self sync.
	^ result at: #localClockOffset! !

!ZTimestampSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 19:43'!
now
	^ ZTimestamp now! !

!ZTimestampSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 14:37'!
ntpEpoch
	^ ZTimestamp year: 1900 month: 1 day: 1! !

!ZTimestampSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 14:47'!
ntpPort
	^ 123! !

!ZTimestampSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 14:36'!
ntpServerAddress
	^ NetNameResolver addressForName: self ntpServerHostname ! !

!ZTimestampSNTPClient methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/25/2013 14:36'!
ntpServerHostname
	^ '0123' atRandom asString, '.pool.ntp.org'! !

!ZTimestampSNTPClient methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 2/25/2013 23:19'!
originateOffset
	^ 25! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/26/2013 10:14'!
receiveDatagram: bytes
	50 timesRepeat: [ 
		((socket receiveUDPDataInto: bytes) at: 1) > 0
			ifTrue: [ ^ self ].
		(Delay forMilliseconds: 5) wait ]! !

!ZTimestampSNTPClient methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 2/25/2013 23:20'!
receiveOffset
	^ 33! !

!ZTimestampSNTPClient methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 2/25/2013 23:19'!
referenceOffset
	^ 17! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/26/2013 09:52'!
remoteClock
	"Return a timstamp with the value of the remote clock,
	more specifically the transmit timestamp of the NTP server"

	| result |
	result := self sync.
	^ result at: #timestamp! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 23:21'!
requestDatagram: now
	| bytes |
	bytes := ByteArray new: 48.
	bytes at: 1 put: 27.	"version 3, client mode"
	((now - self ntpEpoch) asSeconds asByteArrayOfSize: 4)
		withIndexDo: [ :byte :index | bytes at: self transmitOffset + index - 1 put: byte ].
	((now nanoSecond / (10 raisedToInteger: 9) * (2 raisedToInteger: 32)) rounded asByteArrayOfSize: 4) 
		withIndexDo: [ :byte :index | bytes at: self transmitOffset + 4 + index - 1 put: byte ].
	^ bytes! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/26/2013 09:44'!
roundtripDelay
	"Return the round trip delay
	as computed using the NTP clock synchronization algorithm"

	| result |
	result := self sync.
	^ result at: #roundtripDelay! !

!ZTimestampSNTPClient methodsFor: 'operations' stamp: 'SvenVanCaekenberghe 2/25/2013 23:33'!
sync
	"Contact an NTP server and return a processed result,
	a Dictionary holding a #timestamp, #roundTripDelay, #localClockOffset and #datagram"

	| request result t0 t1 |
	t0 := self now.
	request := self requestDatagram: t0.
	result := ByteArray new: 48.
	socket sendUDPData: request toHost: serverAddress port: self ntpPort.
	self receiveDatagram: result.
	t1 := self now.
	^ self interpretResult: result t0: t0 t1: t1! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 23:17'!
timestampAt: offset in: bytes
	| seconds fractions |
	seconds := self fullSecondsAt: offset in: bytes.	
	fractions := self fractionalSecondsAt: offset + 4 in: bytes.
	^ self timestampFromSeconds: seconds fraction: fractions! !

!ZTimestampSNTPClient methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/25/2013 20:25'!
timestampFromSeconds: seconds fraction: fraction
	^ self ntpEpoch + seconds seconds + fraction rounded! !

!ZTimestampSNTPClient methodsFor: 'constants' stamp: 'SvenVanCaekenberghe 2/25/2013 23:20'!
transmitOffset
	^ 41! !

Object subclass: #ZTimezone
	instanceVariableNames: 'id transitions subzones leapSeconds standardIndicators utcIndicators subzonesIndex localTransitions'
	classVariableNames: 'Timezones ZoneInfoLocation ZoneTab'
	poolDictionaries: ''
	category: 'ZTimestamp'!

!ZTimezone commentStamp: 'SvenVanCaekenberghe 3/15/2014 15:36' prior: 0!
I am ZTimezone, representing the timezone information in the standard Olsen database.

  http://en.wikipedia.org/wiki/Tz_database

Usage

You reference timezones by their ID. The list of supported identifiers is accessible using

  ZTimezone timezoneIdentifiers.

To access a timezone do

  ZTimezone id: 'Europe/Brussels'.

The necessary information will be loaded, parsed and cached from a binary file of the zoneinfo database (see also man tzfile). This should work automagically on Mac OS X and Unix, on Windows you have to download the necessary files and specify their location 

  ZTimezone zoneInfoLocation: FileLocator C / 'foo' / 'bar' / 'zoneinfo'.

Once you get a handle on a timezone, the main operation is to query the sub timezone that is applicable at a certain point in time 

  (ZTimezone id: 'Europe/Brussels') subzoneForTimestamp: ZTimestamp now.

The ZSubTimezone instance returned contains information like the UTC offset. Since you'll probably only be interested in that aspect there is a convenience method 

  (ZTimezone id: 'Europe/Brussels') offsetForTimestamp: DateAndTime now.

The flow is that for every GMT timestamp, you get the concrete offset to use for a specific timezone. Note that this is not a constant, it depends on the time periode the timestamp falls in.

The are 2 more convenience methods to quickly convert between GTM and local wall time 

  (ZTimezone id: #'Europe/Brussels') gmtToLocal: ZTimestamp now.
  (ZTimezone id: #'Europe/Brussels') localToGmt: DateAndTime now.

Also note the zoneTab and the timezones are cached in the image. When the TZ database changes, it might be necessary to either call #cleanUp or #reloadAll. When moving images between machines, either all info should be loaded and cached, or it might be necessary to use #zoneInfoLocation: again.

Implementation

A chronological array of transition points in unix time specifies which sub zone is active from that point on to the next.

Limitations

The format 2 data following the format 1 data is not read as it is a duplicate. Leap seconds, the standard/wall indicators and the UTC/local indicators are currently ignored, but their information is read and stored.!

!ZTimezone class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 10:51'!
checkMagicFrom: in
	| version |
	self assert: (in next: 4) = 'TZif' asByteArray.
	version := in next.
	"version 0 is the traditional format.
	version 2 means that the normal data is followed an identical format
	where the offsets are specified using 8 instead of 4 bytes for future proofing"
	in skip: 15! !

!ZTimezone class methodsFor: 'utilties' stamp: 'SvenVanCaekenberghe 3/12/2014 11:21'!
cleanUp
	"Clear our cached data"
	"self cleanUp"
	
	ZoneTab := Timezones := nil! !

!ZTimezone class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/11/2014 00:00'!
count: count do: block
	^ Array new: count streamContents: [ :out |
		count timesRepeat: [ block value: out ] ]! !

!ZTimezone class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 11:18'!
createGMT
	^ ZTimezone new
		id: #GMT;
		transitions: #();
		subzonesIndex: #[0];
		subzones: {(ZSubTimezone gmt)};
		resolveSubzonesIndex;
		yourself! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:12'!
defaultZoneInfoLocation
	"On Mac OS X and Linux/Unix there is a standard location.
	On Windows there is not, so the location has to be specified using #zoneInfoLocation:"

	| standardLocation |
	standardLocation := '/usr/share/zoneinfo/' asFileReference.
	^ standardLocation exists
		ifTrue: [ standardLocation ]
		ifFalse: [ self errorCannotUseDefaultZoneInfoLocation ]! !

!ZTimezone class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 11:10'!
errorCannotUseDefaultZoneInfoLocation
	^ self
		error:
			'Cannot default to standard /usr/share/zoneinfo/ location,' , 
			'please specify a location using #zoneInfoLocation:'! !

!ZTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/12/2014 15:05'!
fromFile: file
	"Parse the binary compiled tzfile at file and return a timezone. Do not cache."
	"self fromFile: '/usr/share/zoneinfo/Europe/Brussels'."

	^ file asFileReference readStreamDo: [ :in |
			in binary. 
			self fromStream: in ]! !

!ZTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/12/2014 11:24'!
fromStream: in
	"Parse binary tzfile format data from in and return a timezone"
	
	| timezone header rawSubzones abbreviations |
	timezone := self new.
	self checkMagicFrom: in.
	header := self readHeaderFrom: in.
	timezone transitions: (self 
		count: (header at: #transitionCount) 
		do: [ :out | out nextPut: in int32 ]).
	timezone subzonesIndex: (in next: (header at: #transitionCount)).
	rawSubzones := self
		count: (header at: #typeCount)
		do: [ :out | out nextPut: { in int32. in next. in next } ].
	timezone leapSeconds: (self
		count: (header at: #leapCount)
		do: [ :out | out nextPut: { in uint32. in uint32 } ]).
	abbreviations := (in next: (header at: #abbreviationCount)) asString. 
	timezone resolveSubzones: rawSubzones abbreviations: abbreviations.
	timezone resolveStandardIndicators: (in next: (header at: #wallCount)).
	timezone resolveUtcIndicators: (in next: (header at: #utcCount)).
	^ timezone! !

!ZTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/12/2014 11:24'!
gmt
	"Return the special GMT timezone, also known as UTC or Zulu"
	
	^ self timezones at: #GMT ifAbsentPut: [ self createGMT ]! !

!ZTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/12/2014 11:19'!
id: id
	"Return the timezone identified by id, either from cache or loading it if necessary"
	
	id = #GMT ifTrue: [ ^ self gmt ].
	^ self timezones
		at: id
		ifAbsentPut: [ self readTimezoneWithId: id ]! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:13'!
locationForId: id
	"Return a FileReference to the file containing the compiled binary data 
	for the timezone identified by id, with $/ indicating directory levels"
	
	| location |
	location := self zoneInfoLocation.
	($/ split: id) do: [ :each |
		location := location / each ].
	^ location! !

!ZTimezone class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/10/2014 23:53'!
readHeaderFrom: in
	| header |
	header := Dictionary new.
	#(#utcCount #wallCount #leapCount #transitionCount #typeCount #abbreviationCount)
		do: [ :each | header at: each put: in uint32 ].
	^ header! !

!ZTimezone class methodsFor: 'instance creation' stamp: 'SvenVanCaekenberghe 3/12/2014 14:15'!
readTimezoneWithId: id
	"Parse and load the timezone with id, do not cache"

	"self readTimezoneWithId: #'Europe/Brussels'."

	| location timezone |
	self zoneTab 
		detect: [ :each | each first = id ] 
		ifNone: [ ^ self error: 'Unknown Timezone ID: ' , id asString ].
	location := self locationForId: id.
	timezone := self fromFile: location.
	timezone id: id.
	^ timezone! !

!ZTimezone class methodsFor: 'utilties' stamp: 'SvenVanCaekenberghe 6/15/2014 17:19'!
readZoneTab
	"Parse and load the zone tab file, do not cache"
	"self readZoneTab"

	| file contents |
	file := self zoneInfoLocation / 'zone.tab'.
	contents := file readStreamDo: [ :in | in upToEnd ].
	^ (contents lines reject: [ :line | line first = $# ])
			collect: [ :lines | 
				| tokens |
				tokens := lines findTokens: Character separators.
				{ tokens third asSymbol. tokens first } ]! !

!ZTimezone class methodsFor: 'utilties' stamp: 'SvenVanCaekenberghe 3/12/2014 11:23'!
reloadAll
	"Clear all cached data and load ALL timezone data. Do cache."
	"self reloadAll"
	
	self cleanUp.
	self timezoneIdentifiers 
		do: [ :each | self id: each ].
	^ self timezones! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:14'!
resetZoneInfoLocation
	"Reset the zone info database location so that the default will be used"
	
	self zoneInfoLocation: nil! !

!ZTimezone class methodsFor: 'utilties' stamp: 'SvenVanCaekenberghe 3/12/2014 11:23'!
statistics
	"Give an indication of the amount of data I currently cache"
	"self statistics."
	
	^ { 
			#totalTimezones -> self timezones size.
			#totalTransitions -> (self timezones sum: [ :each | each transitions size ]).
			#totalSubzones -> (self timezones sum: [ :each | each subzones size ])
		} asDictionary ! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:26'!
timezoneIdentifiers
	"Return the list of supported timezone identifiers.
	#GMT is a special case that is also supported, but it is not in the zonetab."
	
	^ self zoneTab collect: #first! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:26'!
timezones
	"Return the currently cached timezones as a Dictionary"
	
	^ Timezones ifNil: [ Timezones := Dictionary new ]! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:15'!
zoneInfoLocation
	"Return the zone info database location currently in use"
	
	^ ZoneInfoLocation ifNil: [ ZoneInfoLocation := self defaultZoneInfoLocation ]! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:16'!
zoneInfoLocation: directory
	"Set the location of the zone info database to directory"
	
	ZoneInfoLocation := directory asFileReference! !

!ZTimezone class methodsFor: 'acccessing' stamp: 'SvenVanCaekenberghe 3/12/2014 11:16'!
zoneTab
	"Return the zone tab, a list of pairs, timezone identifier and country code"
	
	^ ZoneTab ifNil: [ ZoneTab := self readZoneTab ]! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 10:28'!
byteArrayToBits: byteArray
	| bits |
	bits := 0.
	byteArray withIndexDo: [ :each :index |
		bits := bits bitAt: index put: each ].
	^ bits! !

!ZTimezone methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 3/15/2014 15:32'!
gmtToLocal: timestamp
	"Return a copy of timestamp shifted from GMT to 
	the local wall time as defined by the receiver.
	This works as long as timestamp understands #asUnixTime and #+"

	| offset |
	offset := self offsetForTimestamp: timestamp.
	^ timestamp + offset! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 00:10'!
id
	^ id! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 13:29'!
id: anObject
	id := anObject asSymbol! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
leapSeconds
	^ leapSeconds! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
leapSeconds: anObject
	leapSeconds := anObject! !

!ZTimezone methodsFor: 'convenience' stamp: 'SvenVanCaekenberghe 3/15/2014 15:33'!
localToGmt: timestamp
	"Return a copy of timestamp shifted from the local wall time 
	as defined by the receiver to GMT.
	This works as long as timestamp understands #asUnixTime and #-"

	| offset |
	offset := self offsetForLocalTimestamp: timestamp.
	^ timestamp - offset! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/16/2014 17:52'!
localTransitions
	"Lazily computed transition table in local wall time"

	^ localTransitions
		ifNil: [ 
			localTransitions := self transitions
				withIndexCollect: [ :each :index | 
					each + (self subzoneForTransitionAt: index) offset asSeconds ] ]! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/16/2014 18:36'!
offsetForLocalTimestamp: timestamp
	"Return the offset Duration to be subtracted from timestamp (interpreted as
	local wall time) to get the correct GMT time in the timezone that I represent.
	The offset depends on the point in time (summer/winter time, history).
	All objects that respond to #asUnixTime can serve as a timestamp."
	
	| subzone |
	subzone := self subzoneForLocalTimestamp: timestamp.
	^ subzone offset ! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/13/2014 13:23'!
offsetForTimestamp: timestamp
	"Return the offset Duration to be added to timestamp (interpreted as GMT)
	to get the correct local wall time in the timezone that I represent.
	The offset depends on the point in time (summer/winter time, history).
	All objects that respond to #asUnixTime can serve as a timestamp."
	
	| subzone |
	subzone := self subzoneForTimestamp: timestamp.
	^ subzone offset ! !

!ZTimezone methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:33'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	id ifNotNil: [ stream << id ].
	stream nextPut: $)! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 10:26'!
resolveStandardIndicators: flags
	self standardIndicators: (self byteArrayToBits: flags)! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 10:12'!
resolveSubzones: rawSubzones abbreviations: abbreviations
	subzones := rawSubzones collect: [ :each | 
		| start stop |
		start := each third + 1.
		stop := abbreviations indexOf: Character null startingAt: start.
		ZSubTimezone 
			abbreviation: (abbreviations copyFrom: start to: stop - 1) 
			offset: each first seconds
			dst: each second = 1 ].
	self resolveSubzonesIndex ! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/11/2014 13:02'!
resolveSubzonesIndex
	| resolvedSubzonesIndex |
	resolvedSubzonesIndex := Array new: subzonesIndex size.
	1 to: subzonesIndex size do: [ :each |
		| index |
		index := (subzonesIndex at: each) + 1.
		resolvedSubzonesIndex at: each put: (subzones at: index) ].
	subzonesIndex := resolvedSubzonesIndex ! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/12/2014 10:27'!
resolveUtcIndicators: flags
	self utcIndicators: (self byteArrayToBits: flags)! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/16/2014 18:38'!
resolvedLocalTransitions
	^ self localTransitions
		with: self subzonesIndex
		collect: [ :transition :subzone | (ZTimestamp fromUnixTime: transition) -> subzone ]! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/16/2014 18:38'!
resolvedTransitions
	^ self transitions
		with: self subzonesIndex
		collect: [ :transition :subzone | (ZTimestamp fromUnixTime: transition) -> subzone ]! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
standardIndicators
	^ standardIndicators! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
standardIndicators: anObject
	standardIndicators := anObject! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/16/2014 18:36'!
subzoneForLocalTimestamp: timestamp
	"Return the ZSubTimezone instance for timestamp (interpreted as local wall time)
	to get the correct GMT time in the timezone that I represent.
	The returned object knows the offet Duration to be subtracted, 
	the timezone abbreviation and whether daylight saving was in effect or not.
	The subzone depends on the point in time (summer/winter time, history).
	All objects that respond to #asUnixTime can serve as a timestamp."

	^ self subzoneForLocalUnixTime: timestamp asUnixTime ! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/16/2014 18:40'!
subzoneForLocalUnixTime: unixTime
	"Given unixTime, the number of seconds since 1970-01-01, 
	return the TSubTimezone containing more information,
	applicable to that point in time (interpreted in local wall time)."
	
	^ self localTransitions
		findBinaryIndex: [ :offset | unixTime - offset ]
		do: [ :index | self subzoneForTransitionAt: index ]
		ifNone: [ :lower :upper | 
			| index |
			index := lower ~= 0
				ifTrue: [ lower ]																
				ifFalse: [ upper ].
			self subzoneForTransitionAt: index ]! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/13/2014 13:23'!
subzoneForTimestamp: timestamp
	"Return the ZSubTimezone instance for timestamp (interpreted as GMT)
	to get the correct local wall time in the timezone that I represent.
	The returned object knows the offet Duration to be added, the timezone 
	abbreviation and whether daylight saving was in effect or not.
	The subzone depends on the point in time (summer/winter time, history).
	All objects that respond to #asUnixTime can serve as a timestamp."

	^ self subzoneForUnixTime: timestamp asUnixTime ! !

!ZTimezone methodsFor: 'private' stamp: 'SvenVanCaekenberghe 3/11/2014 10:31'!
subzoneForTransitionAt: index
	^ subzonesIndex at: index! !

!ZTimezone methodsFor: 'public' stamp: 'SvenVanCaekenberghe 3/13/2014 14:12'!
subzoneForUnixTime: unixTime
	"Given unixTime, the number of seconds since 1970-01-01, 
	return the TSubTimezone containing more information,
	applicable to that point in time (interpreted in GMT)."
	
	^ self transitions
		findBinaryIndex: [ :offset | unixTime - offset ]
		do: [ :index | self subzoneForTransitionAt: index ]
		ifNone: [ :lower :upper | 
			| index |
			index := lower ~= 0
				ifTrue: [ lower ]																
				ifFalse: [ upper ].
			self subzoneForTransitionAt: index ]! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
subzones
	^ subzones! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
subzones: anObject
	subzones := anObject! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:31'!
subzonesIndex
	^ subzonesIndex! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/11/2014 10:31'!
subzonesIndex: anObject
	subzonesIndex := anObject! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
transitions
	^ transitions! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
transitions: anObject
	transitions := anObject! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
utcIndicators
	^ utcIndicators! !

!ZTimezone methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 3/10/2014 19:53'!
utcIndicators: anObject
	utcIndicators := anObject! !
ZTimestampFormat initialize!
ZTimestamp initialize!
